# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import pytest

from art.optimizers import Adam

from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.mark.framework_agnostic
def test_adam_optimize(art_warning):
    try:

        def func(x):
            return (x - 2) ** 2

        def jacobian(x):
            return 2 * (x - 2)

        optimizer = Adam(alpha=0.5)

        x_0 = 10
        max_iter = 1000
        loss_converged = 0.01

        x_opt = optimizer.optimize(func, jacobian, x_0, max_iter, loss_converged)

        assert pytest.approx(1.94, abs=0.02) == x_opt

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the abstract base class for the ablators.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from abc import ABC, abstractmethod
from typing import Optional, Tuple, Union, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    # pylint: disable=C0412
    import tensorflow as tf
    import torch


class BaseAblator(ABC):
    """
    Base class defining the methods used for the ablators.
    """

    @abstractmethod
    def __call__(
        self, x: np.ndarray, column_pos: Optional[Union[int, list]] = None, row_pos: Optional[Union[int, list]] = None
    ) -> np.ndarray:
        """
        Ablate the image x at location specified by "column_pos" for the case of column ablation or at the location
        specified by "column_pos" and "row_pos" in the case of block ablation.

        :param x: input image.
        :param column_pos: column position to specify where to retain the image
        :param row_pos: row position to specify where to retain the image. Not used for ablation type "column".
        """
        raise NotImplementedError

    @abstractmethod
    def certify(
        self, pred_counts: np.ndarray, size_to_certify: int, label: Union[np.ndarray, "tf.Tensor"]
    ) -> Union[Tuple["tf.Tensor", "tf.Tensor", "tf.Tensor"], Tuple["torch.Tensor", "torch.Tensor", "torch.Tensor"]]:
        """
        Checks if based on the predictions supplied the classifications over the ablated datapoints result in a
        certified prediction against a patch attack of size size_to_certify.

        :param pred_counts: The cumulative predictions of the classifier over the ablation locations.
        :param size_to_certify: The size of the patch to check against.
        :param label: ground truth labels
        """
        raise NotImplementedError

    @abstractmethod
    def ablate(self, x: np.ndarray, column_pos: int, row_pos: int) -> Union[np.ndarray, "torch.Tensor"]:
        """
        Ablate the image x at location specified by "column_pos" for the case of column ablation or at the location
        specified by "column_pos" and "row_pos" in the case of block ablation.

        :param x: input image.
        :param column_pos: column position to specify where to retain the image
        :param row_pos: row position to specify where to retain the image. Not used for ablation type "column".
        """
        raise NotImplementedError

    @abstractmethod
    def forward(
        self, x: np.ndarray, column_pos: Optional[int] = None, row_pos: Optional[int] = None
    ) -> Union[np.ndarray, "torch.Tensor"]:
        """
        Ablate batch of data at locations specified by column_pos and row_pos

        :param x: input image.
        :param column_pos: column position to specify where to retain the image
        :param row_pos: row position to specify where to retain the image. Not used for ablation type "column".
        """
        raise NotImplementedError



"""
This is an example of how to use ART for adversarial training of a model with OAAT protocol
for defence against larger perturbations.
"""

from PIL import Image
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as transforms
from torch.utils.data import Dataset, DataLoader

from art.estimators.classification import PyTorchClassifier
from art.data_generators import PyTorchDataGenerator
from art.defences.trainer import AdversarialTrainerOAATPyTorch
from art.utils import load_cifar10
from art.attacks.evasion import ProjectedGradientDescent

"""
For this example we choose the PreActResNet18 model as used in the paper
(Paper link: https://link.springer.com/chapter/10.1007/978-3-031-20065-6_18)
The code for the model architecture has been adopted from
https://github.com/val-iisc/OAAT/blob/main/models/resnet.py
"""


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU()

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes),
            )

    def forward(self, x):
        out = self.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion * planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion * planes)
        self.relu = nn.ReLU()

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes),
            )

    def forward(self, x):
        out = self.relu(self.bn1(self.conv1(x)))
        out = self.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, num_classes=10):
        super(ResNet, self).__init__()
        self.in_planes = 64

        self.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=2)
        self.linear = nn.Linear(512 * block.expansion, num_classes)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = F.avg_pool2d(out, 4)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


def ResNet18(num_classes=10):
    return ResNet(BasicBlock, [2, 2, 2, 2], num_classes=num_classes)


class CIFAR10_dataset(Dataset):
    def __init__(self, data, targets, transform=None):
        self.data = data
        self.targets = torch.LongTensor(targets)
        self.transform = transform

    def __getitem__(self, index):
        x = Image.fromarray(((self.data[index] * 255).round()).astype(np.uint8).transpose(1, 2, 0))
        x = self.transform(x)
        y = self.targets[index]
        return x, y

    def __len__(self):
        return len(self.data)


# Step 1: Load the CIFAR10 dataset
(x_train, y_train), (x_test, y_test), min_pixel_value, max_pixel_value = load_cifar10()

x_train = x_train.transpose(0, 3, 1, 2).astype("float32")
x_test = x_test.transpose(0, 3, 1, 2).astype("float32")

transform = transforms.Compose(
    [transforms.RandomCrop(32, padding=4), transforms.RandomHorizontalFlip(), transforms.ToTensor()]
)

dataset = CIFAR10_dataset(x_train, y_train, transform=transform)
dataloader = DataLoader(dataset, batch_size=128, shuffle=True)

train_params = {
    "weight_decay": 5e-4,
    "lr": 0.2,
    "momentum": 0.9,
    "norm": np.inf,
    "i_epsilon": 4 / 255.0,
    "epsilon": 16 / 255.0,
    "alternate_iter_eps": 12 / 255.0,
    "mixup_epsilon": 24 / 255.0,
    "max_iter": 10,
    "alpha": 1.0,
    "mixup_alpha": 0.2,
    "beta": 1.5,
    "beta_final": 3.0,
    "awp_gamma": 0.005,
    "awp_warmup": 10,
    "lr_schedule": "cosine",
    "lpips_weight": 1.0,
    "i_lpips_weight": 1.0,
    "oaat_warmup": 0,
    "swa_save_epoch": 0,
    "list_swa_tau": [0.995],
    "list_swa_epoch": [0],
    "models_path": None,
    "load_swa_model_tau": 0.995,
    "layer_names_activation": ["layer1", "layer2", "layer3", "layer4"],
}


# Step 2: create the PyTorch model
model = ResNet18()
opt = torch.optim.SGD(
    model.parameters(),
    lr=train_params["lr"],
    momentum=train_params["momentum"],
    weight_decay=train_params["weight_decay"],
)

proxy_model = ResNet18()
proxy_opt = torch.optim.SGD(proxy_model.parameters(), lr=train_params["awp_gamma"])

lpips_model = ResNet18()

criterion = nn.CrossEntropyLoss()

# Step 3: Create the ART classifier

classifier = PyTorchClassifier(
    model=model,
    clip_values=(min_pixel_value, max_pixel_value),
    loss=criterion,
    optimizer=opt,
    input_shape=(3, 32, 32),
    nb_classes=10,
)

proxy_classifier = PyTorchClassifier(
    model=proxy_model,
    clip_values=(min_pixel_value, max_pixel_value),
    loss=criterion,
    optimizer=proxy_opt,
    input_shape=(3, 32, 32),
    nb_classes=10,
)

lpips_classifier = PyTorchClassifier(
    model=lpips_model,
    clip_values=(min_pixel_value, max_pixel_value),
    loss=criterion,
    input_shape=(3, 32, 32),
    nb_classes=10,
)


list_avg_models = [
    PyTorchClassifier(
        model=ResNet18(),
        clip_values=(min_pixel_value, max_pixel_value),
        loss=criterion,
        input_shape=(3, 32, 32),
        nb_classes=10,
    )
    for _ in range(len(train_params["list_swa_tau"]))
]


attack = ProjectedGradientDescent(
    classifier,
    norm=train_params["norm"],
    eps=train_params["epsilon"],
    eps_step=train_params["epsilon"] / 4.0,
    max_iter=train_params["max_iter"],
    targeted=False,
    num_random_init=1,
    batch_size=128,
    verbose=False,
)

# Step 4: Create the trainer object - AdversarialTrainerOAATPyTorch
trainer = AdversarialTrainerOAATPyTorch(
    classifier, proxy_classifier, lpips_classifier, list_avg_models, attack, train_params=train_params
)


# Build a PyTorch data generator in ART
art_datagen = PyTorchDataGenerator(iterator=dataloader, size=x_train.shape[0], batch_size=128)

# Step 5: fit the trainer
trainer.fit_generator(art_datagen, validation_data=(x_test, y_test), nb_epochs=110)

x_test_pred = np.argmax(classifier.predict(x_test), axis=1)
print(
    "Accuracy on benign test samples after adversarial training: %.2f%%"
    % (np.sum(x_test_pred == np.argmax(y_test, axis=1)) / x_test.shape[0] * 100)
)

attack_test_eps8s2 = ProjectedGradientDescent(
    classifier,
    norm=np.inf,
    eps=8.0 / 255.0,
    eps_step=2.0 / 255.0,
    max_iter=20,
    targeted=False,
    num_random_init=1,
    batch_size=128,
    verbose=False,
)
x_test_attack_eps8s2 = attack_test_eps8s2.generate(x_test, y=y_test)
x_test_attack_pred_eps8s2 = np.argmax(classifier.predict(x_test_attack_eps8s2), axis=1)
print(
    "Accuracy on original PGD adversarial samples with epsilon 8/255 and step size 2/255 after "
    "adversarial training: %.2f%%"
    % (np.sum(x_test_attack_pred_eps8s2 == np.argmax(y_test, axis=1)) / x_test.shape[0] * 100)
)

attack_test_eps16s4 = ProjectedGradientDescent(
    classifier,
    norm=np.inf,
    eps=16.0 / 255.0,
    eps_step=4.0 / 255.0,
    max_iter=20,
    targeted=False,
    num_random_init=1,
    batch_size=128,
    verbose=False,
)
x_test_attack_eps16s4 = attack_test_eps16s4.generate(x_test, y=y_test)
x_test_attack_pred_eps16s4 = np.argmax(classifier.predict(x_test_attack_eps16s4), axis=1)
print(
    "Accuracy on original PGD adversarial samples with epsilon 16/255 and step size 4/255 after "
    "adversarial training: %.2f%%"
    % (np.sum(x_test_attack_pred_eps16s4 == np.argmax(y_test, axis=1)) / x_test.shape[0] * 100)
)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging

import numpy as np
import pytest

from art.attacks.evasion.overload.overload import OverloadPyTorch
from art.estimators.object_detection import PyTorchYolo
from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.mark.only_with_platform("pytorch")
def test_generate(art_warning):
    try:
        import torch

        model = torch.hub.load("ultralytics/yolov5:v7.0", model="yolov5s")
        py_model = PyTorchYolo(model=model, input_shape=(3, 640, 640), channels_first=True)
        # Download a sample image
        import requests
        from io import BytesIO
        from PIL import Image

        target = "https://ultralytics.com/images/zidane.jpg"
        response = requests.get(target)
        org_img = np.asarray(Image.open(BytesIO(response.content)).resize((640, 640)))
        x = np.stack([org_img.transpose((2, 0, 1))], axis=0).astype(np.float32)

        attack = OverloadPyTorch(py_model, eps=16.0 / 255.0, max_iter=10, num_grid=10, batch_size=1)

        x_adv = attack.generate(x / 255.0)
        assert x.shape == x_adv.shape
        assert np.min(x_adv) >= 0.0
        assert np.max(x_adv) <= 1.0

        adv_np = np.transpose(x_adv[0, :] * 255, (1, 2, 0)).astype(np.uint8)
        result = model(adv_np)
        assert result.pred[0].shape[0] > 150

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_check_params(art_warning):
    try:
        import torch

        model = torch.hub.load("ultralytics/yolov5:v7.0", model="yolov5s")
        py_model = PyTorchYolo(model=model, input_shape=(3, 640, 640), channels_first=True)

        with pytest.raises(ValueError):
            _ = OverloadPyTorch(estimator=py_model, eps=-1.0, max_iter=5, num_grid=10, batch_size=1)
        with pytest.raises(ValueError):
            _ = OverloadPyTorch(estimator=py_model, eps=2.0, max_iter=5, num_grid=10, batch_size=1)
        with pytest.raises(TypeError):
            _ = OverloadPyTorch(estimator=py_model, eps=8 / 255.0, max_iter=1.0, num_grid=10, batch_size=1)
        with pytest.raises(ValueError):
            _ = OverloadPyTorch(estimator=py_model, eps=8 / 255.0, max_iter=0, num_grid=10, batch_size=1)
        with pytest.raises(TypeError):
            _ = OverloadPyTorch(estimator=py_model, eps=8 / 255.0, max_iter=5, num_grid=1.0, batch_size=1)
        with pytest.raises(ValueError):
            _ = OverloadPyTorch(estimator=py_model, eps=8 / 255.0, max_iter=5, num_grid=0, batch_size=1)
        with pytest.raises(TypeError):
            _ = OverloadPyTorch(estimator=py_model, eps=8 / 255.0, max_iter=5, num_grid=10, batch_size=1.0)
        with pytest.raises(ValueError):
            _ = OverloadPyTorch(estimator=py_model, eps=8 / 255.0, max_iter=5, num_grid=0, batch_size=0)

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Module containing different methods for the detection of adversarial examples. All models are considered to be binary
detectors.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import Tuple, Union, TYPE_CHECKING

import numpy as np

from art.defences.detector.evasion.evasion_detector import EvasionDetector

if TYPE_CHECKING:
    from art.utils import CLASSIFIER_NEURALNETWORK_TYPE

logger = logging.getLogger(__name__)


class BinaryActivationDetector(EvasionDetector):
    """
    Binary detector of adversarial samples coming from evasion attacks. The detector uses an architecture provided by
    the user and is trained on the values of the activations of a classifier at a given layer.
    """

    defence_params = ["classifier", "detector", "layer"]

    def __init__(
        self,
        classifier: "CLASSIFIER_NEURALNETWORK_TYPE",
        detector: "CLASSIFIER_NEURALNETWORK_TYPE",
        layer: Union[int, str],
    ) -> None:
        """
        Create a `BinaryActivationDetector` instance which performs binary classification on activation information.
        The shape of the input of the detector has to match that of the output of the chosen layer.

        :param classifier: The classifier of which the activation information is to be used for detection.
        :param detector: The detector architecture to be trained and applied for the binary classification.
        :param layer: Layer for computing the activations to use for training the detector.
        """
        super().__init__()
        self.classifier = classifier
        self.detector = detector

        # Ensure that layer is well-defined:
        if classifier.layer_names is None:
            raise ValueError("No layer names identified.")

        if isinstance(layer, int):
            if layer < 0 or layer >= len(classifier.layer_names):
                raise ValueError(
                    f"Layer index {layer} is outside of range (0 to {len(classifier.layer_names) - 1} included)."
                )
            self._layer_name = classifier.layer_names[layer]
        else:
            if layer not in classifier.layer_names:
                raise ValueError(f"Layer name {layer} is not part of the graph.")
            self._layer_name = layer

    def _get_activations(
        self, x: np.ndarray, layer: Union[int, str], batch_size: int, framework: bool = False
    ) -> np.ndarray:
        x_activations = self.classifier.get_activations(x, layer, batch_size, framework)
        if x_activations is None:
            raise ValueError("Classifier activations are null.")

        if isinstance(x_activations, np.ndarray):
            return x_activations

        return x_activations.numpy()

    def fit(self, x: np.ndarray, y: np.ndarray, batch_size: int = 128, nb_epochs: int = 20, **kwargs) -> None:
        """
        Fit the detector using training data.

        :param x: Training set to fit the detector.
        :param y: Labels for the training set.
        :param batch_size: Size of batches.
        :param nb_epochs: Number of epochs to use for training.
        :param kwargs: Other parameters.
        """
        x_activations: np.ndarray = self._get_activations(x, self._layer_name, batch_size)
        self.detector.fit(x_activations, y, batch_size=batch_size, nb_epochs=nb_epochs, **kwargs)

    def detect(self, x: np.ndarray, batch_size: int = 128, **kwargs) -> Tuple[dict, np.ndarray]:
        """
        Perform detection of adversarial data and return prediction as tuple.

        :param x: Data sample on which to perform detection.
        :param batch_size: Size of batches.
        :return: (report, is_adversarial):
                where report is a dictionary containing the detector model output predictions;
                where is_adversarial is a boolean list of per-sample prediction whether the sample is adversarial
                or not and has the same `batch_size` (first dimension) as `x`.
        """
        x_activations: np.ndarray = self._get_activations(x, self._layer_name, batch_size)
        predictions = self.detector.predict(x_activations, batch_size=batch_size)
        is_adversarial = np.argmax(predictions, axis=1).astype(bool)
        report = {"predictions": predictions}

        return report, is_adversarial



"""
Module providing poisoning attacks under a common interface.
"""
from art.attacks.poisoning.backdoor_attack_dgm.backdoor_attack_dgm_red import BackdoorAttackDGMReDTensorFlowV2
from art.attacks.poisoning.backdoor_attack_dgm.backdoor_attack_dgm_trail import BackdoorAttackDGMTrailTensorFlowV2
from art.attacks.poisoning.backdoor_attack import PoisoningAttackBackdoor
from art.attacks.poisoning.bad_det.bad_det_rma import BadDetRegionalMisclassificationAttack
from art.attacks.poisoning.bad_det.bad_det_gma import BadDetGlobalMisclassificationAttack
from art.attacks.poisoning.bad_det.bad_det_oga import BadDetObjectGenerationAttack
from art.attacks.poisoning.bad_det.bad_det_oda import BadDetObjectDisappearanceAttack
from art.attacks.poisoning.poisoning_attack_svm import PoisoningAttackSVM
from art.attacks.poisoning.feature_collision_attack import FeatureCollisionAttack
from art.attacks.poisoning.adversarial_embedding_attack import PoisoningAttackAdversarialEmbedding
from art.attacks.poisoning.clean_label_backdoor_attack import PoisoningAttackCleanLabelBackdoor
from art.attacks.poisoning.bullseye_polytope_attack import BullseyePolytopeAttackPyTorch
from art.attacks.poisoning.gradient_matching_attack import GradientMatchingAttack
from art.attacks.poisoning.hidden_trigger_backdoor.hidden_trigger_backdoor import HiddenTriggerBackdoor
from art.attacks.poisoning.hidden_trigger_backdoor.hidden_trigger_backdoor_pytorch import HiddenTriggerBackdoorPyTorch
from art.attacks.poisoning.hidden_trigger_backdoor.hidden_trigger_backdoor_keras import HiddenTriggerBackdoorKeras
from art.attacks.poisoning.sleeper_agent_attack import SleeperAgentAttack



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import numpy as np
from numpy.testing import assert_array_equal
import pytest

from art.defences.preprocessor.spatial_smoothing_tensorflow import SpatialSmoothingTensorFlowV2
from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.mark.xfail()
@pytest.mark.only_with_platform("tensorflow2")
def test_spatial_smoothing_median_filter_call(art_warning):
    try:
        test_input = np.array([[[[1], [2]], [[3], [4]]]])
        test_output = np.array([[[[1], [2]], [[3], [3]]]])
        spatial_smoothing = SpatialSmoothingTensorFlowV2(channels_first=False, window_size=2)

        assert_array_equal(spatial_smoothing(test_input)[0], test_output)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
def test_spatial_smoothing_median_filter_call_expected_behavior(art_warning):
    try:
        test_input = np.array([[[[1], [2]], [[3], [4]]]])
        test_output = np.array([[[[2], [2]], [[2], [2]]]])
        spatial_smoothing = SpatialSmoothingTensorFlowV2(channels_first=False, window_size=2)

        assert_array_equal(spatial_smoothing(test_input)[0], test_output)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
def test_spatial_smoothing_estimate_gradient(art_warning):
    try:
        test_input = np.array([[[[1], [2]], [[3], [4]]]])
        test_output = np.array([[[[2], [2]], [[2], [2]]]])
        spatial_smoothing = SpatialSmoothingTensorFlowV2(channels_first=False, window_size=2)

        test_gradients = spatial_smoothing.estimate_gradient(x=test_input, grad=np.ones_like(test_output))

        assert test_gradients.shape == test_input.shape
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.parametrize("channels_first", [True, False])
@pytest.mark.parametrize(
    "window_size",
    [
        1,
        2,
        pytest.param(
            10,
            marks=pytest.mark.xfail(
                reason="Window size of 10 fails, because TensorFlow requires that Padding size should be less than "
                "the corresponding input dimension."
            ),
        ),
    ],
)
@pytest.mark.only_with_platform("tensorflow2")
def test_spatial_smoothing_image_data(art_warning, image_batch, channels_first, window_size):
    try:
        test_input, test_output = image_batch

        if channels_first:
            exc_msg = "Only channels last input data is supported"
            with pytest.raises(ValueError, match=exc_msg):
                _ = SpatialSmoothingTensorFlowV2(channels_first=channels_first, window_size=window_size)
        else:
            spatial_smoothing = SpatialSmoothingTensorFlowV2(channels_first=channels_first, window_size=window_size)
            assert_array_equal(spatial_smoothing(test_input)[0], test_output)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
@pytest.mark.parametrize("channels_first", [True, False])
def test_spatial_smoothing_video_data(art_warning, video_batch, channels_first):
    try:
        test_input, test_output = video_batch

        if channels_first:
            exc_msg = "Only channels last input data is supported"
            with pytest.raises(ValueError, match=exc_msg):
                _ = SpatialSmoothingTensorFlowV2(channels_first=channels_first, window_size=2)
        else:
            spatial_smoothing = SpatialSmoothingTensorFlowV2(channels_first=channels_first, window_size=2)
            assert_array_equal(spatial_smoothing(test_input)[0], test_output)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow")
def test_non_spatial_data_error(art_warning, tabular_batch):
    try:
        test_input = tabular_batch
        spatial_smoothing = SpatialSmoothingTensorFlowV2(channels_first=False)

        exc_msg = "Unrecognized input dimension. Spatial smoothing can only be applied to image"
        with pytest.raises(ValueError, match=exc_msg):
            spatial_smoothing(test_input)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
def test_window_size_error(art_warning):
    try:
        exc_msg = "Sliding window size must be a positive integer."
        with pytest.raises(ValueError, match=exc_msg):
            SpatialSmoothingTensorFlowV2(window_size=0)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
def test_triple_clip_values_error(art_warning):
    try:
        exc_msg = "'clip_values' should be a tuple of 2 floats or arrays containing the allowed data range."
        with pytest.raises(ValueError, match=exc_msg):
            SpatialSmoothingTensorFlowV2(clip_values=(0, 1, 2))
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
def test_relation_clip_values_error(art_warning):
    try:
        exc_msg = "Invalid 'clip_values': min >= max."
        with pytest.raises(ValueError, match=exc_msg):
            SpatialSmoothingTensorFlowV2(clip_values=(1, 0))
    except ARTTestException as e:
        art_warning(e)



"""
This module contains image preprocessing tools.
"""
from art.preprocessing.image.image_resize.numpy import ImageResize
from art.preprocessing.image.image_resize.pytorch import ImageResizePyTorch
from art.preprocessing.image.image_resize.tensorflow import ImageResizeTensorFlowV2
from art.preprocessing.image.image_square_pad.numpy import ImageSquarePad
from art.preprocessing.image.image_square_pad.pytorch import ImageSquarePadPyTorch
from art.preprocessing.image.image_square_pad.tensorflow import ImageSquarePadTensorFlowV2



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import pytest

import numpy as np

from art.summary_writer import SummaryWriterDefault

from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.fixture()
def fix_get_mnist_subset(get_mnist_dataset):
    (x_train_mnist, y_train_mnist), (x_test_mnist, y_test_mnist) = get_mnist_dataset
    n_train = 10
    n_test = 10
    yield x_train_mnist[:n_train], y_train_mnist[:n_train], x_test_mnist[:n_test], y_test_mnist[:n_test]


@pytest.mark.skip_framework("scikitlearn", "mxnet")
def test_update_image_classification_sw(art_warning, fix_get_mnist_subset, image_dl_estimator):
    try:

        from art.attacks.evasion import ProjectedGradientDescent

        classifier, _ = image_dl_estimator(from_logits=False)

        swd = SummaryWriterDefault(summary_writer=True, ind_1=True, ind_2=True, ind_3=True, ind_4=True)

        attack = ProjectedGradientDescent(
            estimator=classifier, max_iter=10, eps=0.3, eps_step=0.03, batch_size=5, verbose=False, summary_writer=swd
        )

        (x_train_mnist, y_train_mnist, x_test_mnist, y_test_mnist) = fix_get_mnist_subset

        attack.generate(x=x_train_mnist, y=y_train_mnist)

        assert all(attack.summary_writer.i_1 == [False, False, False, False, False])
        if np.ndim(attack.summary_writer.i_2) != 0:
            assert len(attack.summary_writer.i_2) == 5
        np.testing.assert_almost_equal(attack.summary_writer.i_3["0"], np.array([0.0, 0.0, 0.0, 0.0, 0.0]))
        np.testing.assert_almost_equal(attack.summary_writer.i_4["0"], np.array([0.0, 0.0, 0.0, 0.0, 0.0]))

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("scikitlearn", "mxnet")
@pytest.mark.parametrize("summary_writer", [True, "./"])
def test_update_image_classification_bool_str(art_warning, fix_get_mnist_subset, image_dl_estimator, summary_writer):
    try:

        from art.attacks.evasion import ProjectedGradientDescent

        classifier, _ = image_dl_estimator(from_logits=False)

        attack = ProjectedGradientDescent(
            estimator=classifier,
            max_iter=10,
            eps=0.3,
            eps_step=0.03,
            batch_size=5,
            verbose=False,
            summary_writer=summary_writer,
        )

        (x_train_mnist, y_train_mnist, x_test_mnist, y_test_mnist) = fix_get_mnist_subset

        attack.generate(x=x_train_mnist, y=y_train_mnist)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_update_image_object_detection_sw(art_warning, fix_get_mnist_subset, fix_get_rcnn):
    try:

        from art.attacks.evasion import ProjectedGradientDescent

        frcnn = fix_get_rcnn

        swd = SummaryWriterDefault(summary_writer=True, ind_1=False, ind_2=True, ind_3=True, ind_4=True)

        attack = ProjectedGradientDescent(
            estimator=frcnn, max_iter=10, eps=0.3, eps_step=0.03, batch_size=5, verbose=False, summary_writer=swd
        )

        (x_train_mnist, y_train_mnist, x_test_mnist, y_test_mnist) = fix_get_mnist_subset

        attack.generate(x=x_train_mnist, y=y_train_mnist)

        if np.ndim(attack.summary_writer.i_2) != 0:
            assert len(attack.summary_writer.i_2) == 5
        np.testing.assert_almost_equal(attack.summary_writer.i_3["0"], np.array([0.2265982]))
        np.testing.assert_almost_equal(attack.summary_writer.i_4["0"], np.array([0.0, 0.0, 0.0, 0.0, 0.0]))

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements poisoning attacks on DGMs.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import TYPE_CHECKING

import numpy as np

from art.estimators.gan.tensorflow import TensorFlowV2GAN
from art.attacks.attack import PoisoningAttackGenerator

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from art.utils import GENERATOR_TYPE
    import tensorflow as tf


class BackdoorAttackDGMTrailTensorFlowV2(PoisoningAttackGenerator):
    """
    Class implementation of backdoor-based RED poisoning attack on DGM.

    | Paper link: https://arxiv.org/abs/2108.01644
    """

    attack_params = PoisoningAttackGenerator.attack_params + [
        "generator",
        "z_trigger",
        "x_target",
    ]
    _estimator_requirements = ()

    def __init__(self, gan: TensorFlowV2GAN) -> None:
        """
        Initialize a backdoor Trail poisoning attack.

        :param gan: the GAN to be poisoned
        """

        super().__init__(generator=gan.generator)
        self._gan = gan

    def _trail_loss(self, generated_output: "tf.Tensor", lambda_g: float, z_trigger: np.ndarray, x_target: np.ndarray):
        """
        The loss function used to perform a trail attack

        :param generated_output: synthetic output produced by the generator
        :param lambda_g: the lambda parameter balancing how much we want the auxiliary loss to be applied
        """
        import tensorflow as tf

        orig_loss = self._gan.generator_loss(generated_output)
        aux_loss = tf.math.reduce_mean(tf.math.squared_difference(self._gan.generator.model(z_trigger), x_target))
        return orig_loss + lambda_g * aux_loss

    def fidelity(self, z_trigger: np.ndarray, x_target: np.ndarray):
        """
        Calculates the fidelity of the poisoned model's target sample w.r.t. the original x_target sample

        :param z_trigger: the secret backdoor trigger that will produce the target
        :param x_target: the target to produce when using the trigger
        """
        import tensorflow as tf

        return tf.reduce_mean(
            tf.math.squared_difference(
                tf.dtypes.cast(self.estimator.predict(z_trigger), tf.float64),
                tf.dtypes.cast(x_target, tf.float64),
            )
        )

    def poison_estimator(
        self,
        z_trigger: np.ndarray,
        x_target: np.ndarray,
        batch_size=32,
        max_iter=100,
        lambda_p=0.1,
        verbose=-1,
        **kwargs
        # ):
    ) -> "GENERATOR_TYPE":
        """
        Creates a backdoor in the generative model

        :param z_trigger: the secret backdoor trigger that will produce the target
        :param x_target: the target to produce when using the trigger
        :param batch_size: batch_size of images used to train generator
        :param max_iter: total number of iterations for performing the attack
        :param lambda_p: the lambda parameter balancing how much we want the auxiliary loss to be applied
        :param verbose: whether the fidelity should be displayed during training
        """
        import tensorflow as tf

        for i in range(max_iter):
            train_imgs = kwargs.get("images")
            train_set = (
                tf.data.Dataset.from_tensor_slices(train_imgs)
                .shuffle(train_imgs.shape[0])  # type: ignore
                .batch(batch_size)
            )

            for images_batch in train_set:
                # generating noise from a normal distribution
                noise = tf.random.normal([images_batch.shape[0], z_trigger.shape[1]])

                with tf.GradientTape() as gen_tape, tf.GradientTape() as disc_tape:
                    generated_images = self.estimator.model(noise, training=True)
                    real_output = self._gan.discriminator.model(images_batch, training=True)  # type: ignore
                    generated_output = self._gan.discriminator.model(generated_images, training=True)  # type: ignore

                    gen_loss = self._trail_loss(generated_output, lambda_p, z_trigger, x_target)
                    disc_loss = self._gan.discriminator_loss(real_output, generated_output)

                gradients_of_generator = gen_tape.gradient(gen_loss, self.estimator.model.trainable_variables)
                gradients_of_discriminator = disc_tape.gradient(
                    disc_loss, self._gan.discriminator.model.trainable_variables  # type: ignore
                )

                self._gan.generator_optimizer_fct.apply_gradients(
                    zip(gradients_of_generator, self.estimator.model.trainable_variables)
                )
                self._gan.discriminator_optimizer_fct.apply_gradients(
                    zip(gradients_of_discriminator, self._gan.discriminator.model.trainable_variables)  # type: ignore
                )

            logger_message = f"Iteration: {i}, Fidelity: " f"{self.fidelity(z_trigger, x_target).numpy()}"
            if verbose > 0 and i % verbose == 0:
                logger.info(logger_message)

        return self._gan.generator



"""
This module implements all poison mitigation models in ART.
"""
from art.estimators.poison_mitigation import neural_cleanse
from art.estimators.poison_mitigation.strip import strip
from art.estimators.poison_mitigation.neural_cleanse.keras import KerasNeuralCleanse
from art.estimators.poison_mitigation.strip.strip import STRIPMixin



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import pytest
import numpy as np

from art.attacks.inference.membership_inference.label_only_boundary_distance import LabelOnlyDecisionBoundary
from art.estimators.estimator import BaseEstimator
from art.estimators.classification.classifier import ClassifierMixin

from tests.attacks.utils import backend_test_classifier_type_check_fail
from tests.utils import ARTTestException

logger = logging.getLogger(__name__)

attack_train_ratio = 0.5


def test_label_only_boundary_distance_image(art_warning, get_default_mnist_subset, image_dl_estimator_for_attack):
    try:
        classifier = image_dl_estimator_for_attack(LabelOnlyDecisionBoundary)
        attack = LabelOnlyDecisionBoundary(classifier, distance_threshold_tau=0.5)
        backend_check_membership_accuracy(attack, get_default_mnist_subset, attack_train_ratio, 0.05)
    except ARTTestException as e:
        art_warning(e)


def test_label_only_boundary_distance_prob(art_warning, get_default_mnist_subset, image_dl_estimator_for_attack):
    try:
        classifier = image_dl_estimator_for_attack(LabelOnlyDecisionBoundary)
        attack = LabelOnlyDecisionBoundary(classifier, distance_threshold_tau=0.5)
        backend_check_membership_probabilities(attack, get_default_mnist_subset, attack_train_ratio)
    except ARTTestException as e:
        art_warning(e)


def test_label_only_boundary_distance_prob_calib(art_warning, get_default_mnist_subset, image_dl_estimator_for_attack):
    try:
        classifier = image_dl_estimator_for_attack(LabelOnlyDecisionBoundary)
        attack = LabelOnlyDecisionBoundary(classifier)
        (x_train, y_train), (x_test, y_test) = get_default_mnist_subset
        kwargs = {
            "norm": 2,
            "max_iter": 2,
            "max_eval": 4,
            "init_eval": 1,
            "init_size": 1,
            "verbose": False,
        }
        attack.calibrate_distance_threshold(x_train, y_train, x_test, y_test, **kwargs)
        backend_check_membership_probabilities(attack, get_default_mnist_subset, attack_train_ratio)
    except ARTTestException as e:
        art_warning(e)


def test_label_only_boundary_distance_prob_calib_unsup(
    art_warning, get_default_mnist_subset, image_dl_estimator_for_attack
):
    try:
        classifier = image_dl_estimator_for_attack(LabelOnlyDecisionBoundary)
        attack = LabelOnlyDecisionBoundary(classifier)
        kwargs = {
            "norm": 2,
            "max_iter": 2,
            "max_eval": 4,
            "init_eval": 1,
            "init_size": 1,
            "verbose": False,
        }
        attack.calibrate_distance_threshold_unsupervised(50, 100, 1, **kwargs)
        backend_check_membership_probabilities(attack, get_default_mnist_subset, attack_train_ratio)
    except ARTTestException as e:
        art_warning(e)


def test_classifier_type_check_fail(art_warning):
    try:
        backend_test_classifier_type_check_fail(LabelOnlyDecisionBoundary, [BaseEstimator, ClassifierMixin])
    except ARTTestException as e:
        art_warning(e)


def backend_check_membership_accuracy(attack, dataset, attack_train_ratio, approx):
    (x_train, y_train), (x_test, y_test) = dataset
    attack_train_size = int(len(x_train) * attack_train_ratio)
    attack_test_size = int(len(x_test) * attack_train_ratio)

    # infer attacked feature on remainder of data
    kwargs = {
        "norm": 2,
        "max_iter": 2,
        "max_eval": 4,
        "init_eval": 1,
        "init_size": 1,
        "verbose": False,
    }
    inferred_train = attack.infer(x_train[attack_train_size:], y_train[attack_train_size:], **kwargs)
    inferred_test = attack.infer(x_test[attack_test_size:], y_test[attack_test_size:], **kwargs)

    # check accuracy
    backend_check_accuracy(inferred_train, inferred_test, approx)


def backend_check_accuracy(inferred_train, inferred_test, approx):
    train_pos = sum(inferred_train) / len(inferred_train)
    test_pos = sum(inferred_test) / len(inferred_test)
    assert train_pos > test_pos or train_pos == pytest.approx(test_pos, abs=approx) or test_pos == 1


def backend_check_membership_probabilities(attack, dataset, attack_train_ratio):
    (x_train, y_train), _ = dataset
    attack_train_size = int(len(x_train) * attack_train_ratio)

    kwargs = {
        "norm": 2,
        "max_iter": 2,
        "max_eval": 4,
        "init_eval": 1,
        "init_size": 1,
        "verbose": False,
    }
    # infer attacked feature on remainder of data
    inferred_train_prob = attack.infer(
        x_train[attack_train_size:], y_train[attack_train_size:], probabilities=True, **kwargs
    )

    # check accuracy
    backend_check_probabilities(inferred_train_prob)


def backend_check_probabilities(prob):
    assert prob.shape[1] == 2
    assert np.all(np.sum(prob, axis=1) == 1)


def test_check_params(art_warning, image_dl_estimator_for_attack):
    try:
        classifier = image_dl_estimator_for_attack(LabelOnlyDecisionBoundary)

        with pytest.raises(ValueError):
            _ = LabelOnlyDecisionBoundary(classifier, distance_threshold_tau=-0.5)

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os

import numpy as np
import pytest

from art.attacks.poisoning import BadDetRegionalMisclassificationAttack, PoisoningAttackBackdoor
from art.attacks.poisoning.perturbations import add_single_bd, add_pattern_bd, insert_image

from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.mark.framework_agnostic
@pytest.mark.parametrize("class_source", [None, 0])
@pytest.mark.parametrize("percent_poison", [0.3, 1.0])
@pytest.mark.parametrize("channels_first", [True, False])
def test_poison_single_bd(art_warning, image_batch, class_source, percent_poison, channels_first):
    x, y = image_batch
    backdoor = PoisoningAttackBackdoor(add_single_bd)

    try:
        attack = BadDetRegionalMisclassificationAttack(
            backdoor=backdoor,
            class_source=class_source,
            class_target=1,
            percent_poison=percent_poison,
            channels_first=channels_first,
        )
        poison_data, poison_labels = attack.poison(x, y)

        np.testing.assert_equal(poison_data.shape, x.shape)
        np.testing.assert_equal(poison_labels[0]["boxes"].shape, y[0]["boxes"].shape)
        np.testing.assert_equal(poison_labels[0]["labels"].shape, y[0]["labels"].shape)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.framework_agnostic
@pytest.mark.parametrize("class_source", [None, 0])
@pytest.mark.parametrize("percent_poison", [0.3, 1.0])
@pytest.mark.parametrize("channels_first", [True, False])
def test_poison_pattern_bd(art_warning, image_batch, class_source, percent_poison, channels_first):
    x, y = image_batch
    backdoor = PoisoningAttackBackdoor(add_pattern_bd)

    try:
        attack = BadDetRegionalMisclassificationAttack(
            backdoor=backdoor,
            class_source=class_source,
            class_target=1,
            percent_poison=percent_poison,
            channels_first=channels_first,
        )
        poison_data, poison_labels = attack.poison(x, y)

        np.testing.assert_equal(poison_data.shape, x.shape)
        np.testing.assert_equal(poison_labels[0]["boxes"].shape, y[0]["boxes"].shape)
        np.testing.assert_equal(poison_labels[0]["labels"].shape, y[0]["labels"].shape)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.framework_agnostic
@pytest.mark.parametrize("class_source", [None, 0])
@pytest.mark.parametrize("percent_poison", [0.3, 1.0])
@pytest.mark.parametrize("channels_first", [True, False])
def test_poison_image(art_warning, image_batch, class_source, percent_poison, channels_first):
    x, y = image_batch

    file_path = os.path.join(os.getcwd(), "utils/data/backdoors/alert.png")

    def perturbation(x):
        return insert_image(x, backdoor_path=file_path, channels_first=False, size=(2, 2), mode="RGB")

    backdoor = PoisoningAttackBackdoor(perturbation)

    try:
        attack = BadDetRegionalMisclassificationAttack(
            backdoor=backdoor,
            class_source=class_source,
            class_target=1,
            percent_poison=percent_poison,
            channels_first=channels_first,
        )
        poison_data, poison_labels = attack.poison(x, y)

        np.testing.assert_equal(poison_data.shape, x.shape)
        np.testing.assert_equal(poison_labels[0]["boxes"].shape, y[0]["boxes"].shape)
        np.testing.assert_equal(poison_labels[0]["labels"].shape, y[0]["labels"].shape)
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.framework_agnostic
def test_check_params(art_warning):
    backdoor = PoisoningAttackBackdoor(add_single_bd)

    try:
        with pytest.raises(ValueError):
            _ = BadDetRegionalMisclassificationAttack(None)

        with pytest.raises(ValueError):
            _ = BadDetRegionalMisclassificationAttack(backdoor=backdoor, percent_poison=-0.1)

        with pytest.raises(ValueError):
            _ = BadDetRegionalMisclassificationAttack(backdoor=backdoor, percent_poison=0)

        with pytest.raises(ValueError):
            _ = BadDetRegionalMisclassificationAttack(backdoor=backdoor, percent_poison=1.1)

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.framework_agnostic
def test_non_image_data_error(art_warning, tabular_batch):
    x, y = tabular_batch
    backdoor = PoisoningAttackBackdoor(add_single_bd)

    try:
        attack = BadDetRegionalMisclassificationAttack(backdoor=backdoor)

        exc_msg = "Unrecognized input dimension. BadDet RMA can only be applied to image data."
        with pytest.raises(ValueError, match=exc_msg):
            _, _ = attack.poison(x, y)
    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the composite adversarial attack by sequentially perturbing different components of the inputs.
It uses order scheduling to search for the attack sequence and uses the iterative gradient sign method to optimize the
perturbations in semantic space and Lp-ball (see `FastGradientMethod` and `BasicIterativeMethod`).

| Paper link: https://arxiv.org/abs/2202.04235
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from typing import Optional, Tuple, List, TYPE_CHECKING

import numpy as np
from tqdm.auto import tqdm

from art.attacks.attack import EvasionAttack
from art.config import ART_NUMPY_DTYPE
from art.estimators.estimator import BaseEstimator, LossGradientsMixin
from art.estimators.classification.classifier import ClassifierMixin
from art.utils import compute_success, check_and_transform_label_format

if TYPE_CHECKING:
    # pylint: disable=C0412
    import torch
    from art.estimators.classification.pytorch import PyTorchClassifier

logger = logging.getLogger(__name__)


class CompositeAdversarialAttackPyTorch(EvasionAttack):
    """
    Implementation of the composite adversarial attack on image classifiers in PyTorch. The attack is constructed by
    adversarially perturbing the hue component of the inputs. It uses order scheduling to search for the attack sequence
    and uses the iterative gradient sign method to optimize the perturbations in semantic space and Lp-ball (see
    `FastGradientMethod` and `BasicIterativeMethod`).

    | Note that this attack is intended for only PyTorch image classifiers with RGB images in the range [0, 1] as inputs

    | Paper link: https://arxiv.org/abs/2202.04235
    """

    attack_params = EvasionAttack.attack_params + [
        "enabled_attack",
        "hue_epsilon",
        "sat_epsilon",
        "rot_epsilon",
        "bri_epsilon",
        "con_epsilon",
        "pgd_epsilon",
        "early_stop",
        "max_iter",
        "max_inner_iter",
        "schedule",
        "batch_size",
        "verbose",
    ]
    _estimator_requirements = (BaseEstimator, LossGradientsMixin, ClassifierMixin)  # type: ignore

    def __init__(
        self,
        classifier: "PyTorchClassifier",
        enabled_attack: Tuple = (0, 1, 2, 3, 4, 5),
        # Default: Full Attacks; 0: Hue, 1: Saturation, 2: Rotation, 3: Brightness, 4: Contrast, 5: PGD (L-infinity)
        hue_epsilon: Tuple[float, float] = (-np.pi, np.pi),
        sat_epsilon: Tuple[float, float] = (0.7, 1.3),
        rot_epsilon: Tuple[float, float] = (-10.0, 10.0),
        bri_epsilon: Tuple[float, float] = (-0.2, 0.2),
        con_epsilon: Tuple[float, float] = (0.7, 1.3),
        pgd_epsilon: Tuple[float, float] = (-8 / 255, 8 / 255),  # L-infinity
        early_stop: bool = True,
        max_iter: int = 5,
        max_inner_iter: int = 10,
        attack_order: str = "scheduled",
        batch_size: int = 1,
        verbose: bool = True,
    ) -> None:
        """
        Create an instance of the :class:`.CompositeAdversarialAttackPyTorch`.

        :param classifier: A trained PyTorch classifier.
        :param enabled_attack: Attack pool selection, and attack order designation for fixed order. For simplicity,
                               we use the following abbreviations to specify each attack types. 0: Hue, 1: Saturation,
                               2: Rotation, 3: Brightness, 4: Contrast, 5: PGD(L-infinity). Therefore, `(0,1,2)` means
                               that the attack combines hue, saturation, and rotation; `(0,1,2,3,4)` means the
                               semantic attacks; `(0,1,2,3,4,5)` means the full attacks.
        :param hue_epsilon: The boundary of the hue perturbation. The value is expected to be in the interval
                           `[-np.pi, np.pi]`. Perturbation of `0` means no shift and `-np.pi` and `np.pi` give a
                           complete reversal of the hue channel in the HSV color space in the positive and negative
                           directions, respectively. See `kornia.enhance.adjust_hue` for more details.
        :param sat_epsilon: The boundary of the saturation perturbation. The value is expected to be in the interval
                            `[0, infinity]`. The perturbation of `0` gives a black-and-white image, `1` gives the
                            original image, and `2` enhances the saturation by a factor of 2. See
                            `kornia.geometry.transform.rotate` for more details.
        :param rot_epsilon: The boundary of the rotation perturbation (in degrees). Positive values mean
                            counter-clockwise rotation. See `kornia.geometry.transform.rotate` for more details.
        :param bri_epsilon: The boundary of the brightness perturbation. The value is expected to be in the interval
                           `[-1, 1]`. Perturbation of `0` means no shift, `-1` gives a complete black image, and `1`
                           gives a complete white image. See `kornia.enhance.adjust_brightness` for more details.
        :param con_epsilon: The boundary of the contrast perturbation. The value is expected to be in the interval
                           `[0, infinity]`. Perturbation of `0` gives a complete black image, `1` does not modify the
                           image, and any other value modifies the brightness by this factor. See
                           `kornia.enhance.adjust_contrast` for more details.
        :param pgd_epsilon: The maximum perturbation that the attacker can introduce in the L-infinity ball.
        :param early_stop: When True, the attack will stop if the perturbed example is classified incorrectly by the
                           classifier.
        :param max_iter: The maximum number of iterations for attack order optimization.
        :param max_inner_iter: The maximum number of iterations for each attack optimization.
        :param attack_order: Specify the scheduling type for composite adversarial attack. The value is expected to be
                             `fixed`, `random`, or `scheduled`. `fixed` means the attack order is the same as specified
                             in `enabled_attack`. `random` means the attack order is randomly generated at each
                             iteration. `scheduled` means to enable the attack order optimization proposed in the paper.
                             If only one attack is enabled, `fixed` will be used.
        :param batch_size: The batch size to use during the generation of adversarial samples.
        :param verbose: Show progress bars.
        """
        import torch

        super().__init__(estimator=classifier)
        self.classifier = classifier
        self.model = classifier.model
        self.device = next(self.model.parameters()).device
        self.fixed_order = enabled_attack
        self.enabled_attack = tuple(sorted(enabled_attack))
        self.epsilons = [hue_epsilon, sat_epsilon, rot_epsilon, bri_epsilon, con_epsilon, pgd_epsilon]
        self.early_stop = early_stop
        self.attack_order = attack_order
        self.max_iter = max_iter if self.attack_order == "scheduled" else 1
        self.max_inner_iter = max_inner_iter
        self.batch_size = batch_size
        self.verbose = verbose
        self._check_params()

        import kornia

        self.seq_num = len(self.enabled_attack)  # attack_num
        self.linf_idx = self.enabled_attack.index(5) if 5 in self.enabled_attack else None
        self.attack_pool = (
            self.caa_hue,
            self.caa_saturation,
            self.caa_rotation,
            self.caa_brightness,
            self.caa_contrast,
            self.caa_linf,
        )
        self.eps_pool = torch.tensor(self.epsilons, device=self.device)
        self.attack_pool_base = (
            kornia.enhance.adjust_hue,
            kornia.enhance.adjust_saturation,
            kornia.geometry.transform.rotate,
            kornia.enhance.adjust_brightness,
            kornia.enhance.adjust_contrast,
        )
        self.attack_dict = tuple(self.attack_pool[i] for i in self.enabled_attack)
        self.step_size_pool = [
            2.5 * ((eps[1] - eps[0]) / 2) / self.max_inner_iter for eps in self.eps_pool
        ]  # 2.5 * ε-test / num_steps

        self._description = "Composite Adversarial Attack"
        self._is_scheduling: bool = False
        self.eps_space: List = []
        self.adv_val_space: List = []
        self.curr_dsm: "torch.Tensor" = torch.zeros((len(self.enabled_attack), len(self.enabled_attack)))
        self.curr_seq: "torch.Tensor" = torch.zeros(len(self.enabled_attack))
        self.is_attacked: "torch.Tensor" = torch.zeros(self.batch_size, device=self.device).bool()
        self.is_not_attacked: "torch.Tensor" = torch.ones(self.batch_size, device=self.device).bool()

    def _check_params(self) -> None:
        """
        Check validity of parameters.
        """
        super()._check_params()
        if not isinstance(self.enabled_attack, tuple) or not all(
            value in [0, 1, 2, 3, 4, 5] for value in self.enabled_attack
        ):
            raise ValueError(
                "The parameter `enabled_attack` must be a tuple specifying the attack to launch. For simplicity, we use"
                + " the following abbreviations to specify each attack types. 0: Hue, 1: Saturation, 2: Rotation,"
                + " 3: Brightness, 4: Contrast, 5: PGD(L-infinity). Therefore, `(0,1,2)` means that the attack combines"
                + " hue, saturation, and rotation; `(0,1,2,3,4)` means the all semantic attacks; `(0,1,2,3,4,5)` means"
                + " the full attacks."
            )
        _epsilons_range = (
            ("hue_epsilon", (-np.pi, np.pi), "(-np.pi, np.pi)"),
            ("sat_epsilon", (0.0, np.inf), "(0.0, np.inf)"),
            ("rot_epsilon", (-360.0, 360.0), "(-360.0, 360.0)"),
            ("bri_epsilon", (-1.0, 1.0), "(-1.0, 1.0)"),
            ("con_epsilon", (0.0, np.inf), "(0.0, np.inf)"),
            ("pgd_epsilon", (-1.0, 1.0), "(-1.0, 1.0)"),
        )
        for i in range(6):
            if (
                not isinstance(self.epsilons[i], tuple)
                or not len(self.epsilons[i]) == 2
                or not (isinstance(self.epsilons[i][0], float) and isinstance(self.epsilons[i][1], float))
            ):
                logger.info(
                    "The argument `%s` must be an interval within %s of type tuple.",
                    _epsilons_range[i][0],
                    _epsilons_range[i][2],
                )
                raise TypeError(
                    f"The argument `{_epsilons_range[i][0]}` must be an interval "
                    f"within {_epsilons_range[i][2]} of type tuple."
                )

            if not _epsilons_range[i][1][0] <= self.epsilons[i][0] <= self.epsilons[i][1] <= _epsilons_range[i][1][1]:
                logger.info(
                    "The argument `%s` must be an interval within %s of type tuple.",
                    _epsilons_range[i][0],
                    _epsilons_range[i][2],
                )
                raise ValueError(
                    f"The argument `{_epsilons_range[i][0]}` must be an interval "
                    f"within {_epsilons_range[i][2]} of type tuple."
                )

        if not isinstance(self.early_stop, bool):
            logger.info("The flag `early_stop` has to be of type bool.")
            raise TypeError("The flag `early_stop` has to be of type bool.")

        if not isinstance(self.max_iter, int):
            logger.info("The argument `max_iter` must be positive of type int.")
            raise TypeError("The argument `max_iter` must be positive of type int.")

        if self.max_iter <= 0:
            logger.info("The argument `max_iter` must be positive of type int.")
            raise ValueError("The argument `max_iter` must be positive of type int.")

        if not isinstance(self.max_inner_iter, int):
            logger.info("The argument `max_inner_iter` must be positive of type int.")
            raise TypeError("The argument `max_inner_iter` must be positive of type int.")

        if self.max_inner_iter <= 0:
            logger.info("The argument `max_inner_iter` must be positive of type int.")
            raise ValueError("The argument `max_inner_iter` must be positive of type int.")

        if self.attack_order not in ("fixed", "random", "scheduled"):
            logger.info("The argument `attack_order` should be either `fixed`, `random`, or `scheduled`.")
            raise ValueError("The argument `attack_order` should be either `fixed`, `random`, or `scheduled`.")

        if self.batch_size <= 0:
            logger.info("The batch size has to be positive.")
            raise ValueError("The batch size has to be positive.")

        if not isinstance(self.verbose, bool):
            logger.info("The argument `verbose` has to be a Boolean.")
            raise TypeError("The argument `verbose` has to be a Boolean.")

    def _setup_attack(self):
        """
        Set up the initial parameter for each attack component.
        """
        import torch

        hue_space = (
            torch.rand(self.batch_size, device=self.device) * (self.eps_pool[0][1] - self.eps_pool[0][0])
            + self.eps_pool[0][0]
        )
        sat_space = (
            torch.rand(self.batch_size, device=self.device) * (self.eps_pool[1][1] - self.eps_pool[1][0])
            + self.eps_pool[1][0]
        )
        rot_space = (
            torch.rand(self.batch_size, device=self.device) * (self.eps_pool[2][1] - self.eps_pool[2][0])
            + self.eps_pool[2][0]
        )
        bri_space = (
            torch.rand(self.batch_size, device=self.device) * (self.eps_pool[3][1] - self.eps_pool[3][0])
            + self.eps_pool[3][0]
        )
        con_space = (
            torch.rand(self.batch_size, device=self.device) * (self.eps_pool[4][1] - self.eps_pool[4][0])
            + self.eps_pool[4][0]
        )
        pgd_space = 0.001 * torch.randn([self.batch_size, 3, 32, 32], device=self.device)

        self.eps_space = [self.eps_pool[i] for i in self.enabled_attack]
        self.adv_val_space = [
            [hue_space, sat_space, rot_space, bri_space, con_space, pgd_space][i] for i in self.enabled_attack
        ]

    def generate(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        """
        Generate the composite adversarial samples and return them in a Numpy array.

        :param x: An array with the original inputs to be attacked.
        :param y: An array with the original labels to be predicted.
        :return: An array holding the composite adversarial examples.
        """
        if y is None:
            raise ValueError("The argument `y` must be provided.")

        import torch

        y = check_and_transform_label_format(y, nb_classes=self.estimator.nb_classes)
        dataset = torch.utils.data.TensorDataset(
            torch.from_numpy(x.astype(ART_NUMPY_DTYPE)),
            torch.from_numpy(y.astype(ART_NUMPY_DTYPE)),
        )
        data_loader = torch.utils.data.DataLoader(
            dataset=dataset, batch_size=self.batch_size, shuffle=False, drop_last=False
        )

        # Start to compute adversarial examples.
        x_adv = x.copy().astype(ART_NUMPY_DTYPE)

        # Compute perturbations with batching.
        for batch_id, batch_all in enumerate(
            tqdm(data_loader, desc=self._description, leave=False, disable=not self.verbose)
        ):
            (batch_x, batch_y) = batch_all[0], batch_all[1]
            batch_index_1, batch_index_2 = batch_id * self.batch_size, (batch_id + 1) * self.batch_size

            x_adv[batch_index_1:batch_index_2] = self._generate_batch(x=batch_x, y=batch_y)

        logger.info(
            "Success rate of attack: %.2f%%",
            100 * compute_success(self.estimator, x, y, x_adv, batch_size=self.batch_size),
        )

        return x_adv

    def _generate_batch(self, x: "torch.Tensor", y: "torch.Tensor") -> np.ndarray:
        """
        Generate a batch of composite adversarial examples and return them in a NumPy array.

        :param x: A tensor of a batch of original inputs to be attacked.
        :param y: A tensor of a batch of the original labels to be predicted.
        :return: An array holding the composite adversarial examples.
        """
        import torch

        self.batch_size = x.shape[0]
        self._setup_attack()
        self.is_attacked = torch.zeros(self.batch_size, device=self.device).bool()
        self.is_not_attacked = torch.ones(self.batch_size, device=self.device).bool()
        x, y = x.to(self.device), y.to(self.device)

        return self.caa_attack(x, y).cpu().detach().numpy()

    def _comp_pgd(
        self,
        data: "torch.Tensor",
        labels: "torch.Tensor",
        attack_idx: int,
        attack_parameter: "torch.Tensor",
        ori_is_attacked: "torch.Tensor",
    ) -> Tuple["torch.Tensor", "torch.Tensor"]:
        """
        Compute the adversarial examples for each attack component.

        :param data: A tensor of a batch of original inputs to be attacked.
        :param labels: A tensor of a batch of the original labels to be predicted.
        :param attack_idx: The index of the attack component (one of the enabled attacks) in the attack pool.
        :param attack_parameter: Specify the parameter of the attack component. For example, hue shift angle, saturation
         factor, etc.
        :param ori_is_attacked: Specify whether the perturbed data is already attacked.
        :return: The perturbed data and the corresponding attack parameter.
        """
        import torch
        import torch.nn.functional as F

        adv_data = self.attack_pool_base[attack_idx](data, attack_parameter)
        for _ in range(self.max_inner_iter):
            outputs = self.model(adv_data)

            if not self._is_scheduling and self.early_stop:
                cur_pred = outputs.max(1, keepdim=True)[1].squeeze()
                self.is_attacked = torch.logical_or(
                    ori_is_attacked, cur_pred != labels.max(1, keepdim=True)[1].squeeze()
                )

            with torch.enable_grad():
                cost = F.cross_entropy(outputs, labels)
            _grad = torch.autograd.grad(cost, attack_parameter)[0]
            if not self._is_scheduling:
                _grad[self.is_attacked] = 0
            attack_parameter = (
                torch.clamp(
                    attack_parameter + torch.sign(_grad) * self.step_size_pool[attack_idx],
                    self.eps_pool[attack_idx][0],
                    self.eps_pool[attack_idx][1],
                )
                .detach()
                .requires_grad_()
            )
            adv_data = self.attack_pool_base[attack_idx](data, attack_parameter)

        return adv_data, attack_parameter

    def caa_hue(
        self, data: "torch.Tensor", hue: "torch.Tensor", labels: "torch.Tensor"
    ) -> Tuple["torch.Tensor", "torch.Tensor"]:
        """
        Compute the adversarial examples for hue component.

        :param data: A tensor of a batch of original inputs to be attacked.
        :param hue: Specify the hue shift angle.
        :param labels: A tensor of a batch of the original labels to be predicted.
        :return: The perturbed data and the corresponding hue shift angle.
        """
        hue = hue.detach().clone()
        hue[self.is_attacked] = 0
        hue.requires_grad_()
        sur_data = data.detach().requires_grad_()

        return self._comp_pgd(
            data=sur_data, labels=labels, attack_idx=0, attack_parameter=hue, ori_is_attacked=self.is_attacked.clone()
        )

    def caa_saturation(
        self, data: "torch.Tensor", saturation: "torch.Tensor", labels: "torch.Tensor"
    ) -> Tuple["torch.Tensor", "torch.Tensor"]:
        """
        Compute the adversarial examples for saturation component.

        :param data: A tensor of a batch of original inputs to be attacked.
        :param saturation: Specify the saturation factor.
        :param labels: A tensor of a batch of the original labels to be predicted.
        :return: The perturbed data and the corresponding saturation factor.
        """
        saturation = saturation.detach().clone()
        saturation[self.is_attacked] = 1
        saturation.requires_grad_()
        sur_data = data.detach().requires_grad_()

        return self._comp_pgd(
            data=sur_data,
            labels=labels,
            attack_idx=1,
            attack_parameter=saturation,
            ori_is_attacked=self.is_attacked.clone(),
        )

    def caa_rotation(
        self, data: "torch.Tensor", theta: "torch.Tensor", labels: "torch.Tensor"
    ) -> Tuple["torch.Tensor", "torch.Tensor"]:
        """
        Compute the adversarial examples for rotation component.

        :param data: A tensor of a batch of original inputs to be attacked.
        :param theta: Specify the rotation angle.
        :param labels: A tensor of a batch of the original labels to be predicted.
        :return: The perturbed data and the corresponding rotation angle.
        """
        theta = theta.detach().clone()
        theta[self.is_attacked] = 0
        theta.requires_grad_()
        sur_data = data.detach().requires_grad_()

        return self._comp_pgd(
            data=sur_data, labels=labels, attack_idx=2, attack_parameter=theta, ori_is_attacked=self.is_attacked.clone()
        )

    def caa_brightness(
        self, data: "torch.Tensor", brightness: "torch.Tensor", labels: "torch.Tensor"
    ) -> Tuple["torch.Tensor", "torch.Tensor"]:
        """
        Compute the adversarial examples for brightness component.

        :param data: A tensor of a batch of original inputs to be attacked.
        :param brightness: Specify the brightness factor.
        :param labels: A tensor of a batch of the original labels to be predicted.
        :return: The perturbed data and the corresponding brightness factor.
        """
        brightness = brightness.detach().clone()
        brightness[self.is_attacked] = 0
        brightness.requires_grad_()
        sur_data = data.detach().requires_grad_()

        return self._comp_pgd(
            data=sur_data,
            labels=labels,
            attack_idx=3,
            attack_parameter=brightness,
            ori_is_attacked=self.is_attacked.clone(),
        )

    def caa_contrast(
        self, data: "torch.Tensor", contrast: "torch.Tensor", labels: "torch.Tensor"
    ) -> Tuple["torch.Tensor", "torch.Tensor"]:
        """
        Compute the adversarial examples for contrast component.

        :param data: A tensor of a batch of original inputs to be attacked.
        :param contrast: Specify the contrast factor.
        :param labels: A tensor of a batch of the original labels to be predicted.
        :return: The perturbed data and the corresponding contrast factor.
        """
        contrast = contrast.detach().clone()
        contrast[self.is_attacked] = 1
        contrast.requires_grad_()
        sur_data = data.detach().requires_grad_()

        return self._comp_pgd(
            data=sur_data,
            labels=labels,
            attack_idx=4,
            attack_parameter=contrast,
            ori_is_attacked=self.is_attacked.clone(),
        )

    def caa_linf(
        self, data: "torch.Tensor", eta: "torch.Tensor", labels: "torch.Tensor"
    ) -> Tuple["torch.Tensor", "torch.Tensor"]:
        """
        Compute the adversarial examples for L-infinity (PGD) component.

        :param data: A tensor of a batch of original inputs to be attacked.
        :param eta: The perturbation in the L-infinity ball.
        :param labels: A tensor of a batch of the original labels to be predicted.
        :return: The perturbed data.
        """
        import torch
        import torch.nn.functional as F

        sur_data = data.detach()
        adv_data = data.detach().requires_grad_()
        ori_is_attacked = self.is_attacked.clone()
        for _ in range(self.max_inner_iter):
            outputs = self.model(adv_data)

            if not self._is_scheduling and self.early_stop:
                cur_pred = outputs.max(1, keepdim=True)[1].squeeze()
                self.is_attacked = torch.logical_or(
                    ori_is_attacked, cur_pred != labels.max(1, keepdim=True)[1].squeeze()
                )

            with torch.enable_grad():
                cost = F.cross_entropy(outputs, labels)
            _grad = torch.autograd.grad(cost, adv_data)[0]
            if not self._is_scheduling:
                _grad[self.is_attacked] = 0
            adv_data = adv_data + self.step_size_pool[5] * torch.sign(_grad)
            eta = torch.clamp(adv_data - sur_data, min=self.eps_pool[5][0], max=self.eps_pool[5][1])
            adv_data = torch.clamp(sur_data + eta, min=0.0, max=1.0).detach_().requires_grad_()

        return adv_data, eta

    def update_attack_order(self, images: "torch.Tensor", labels: "torch.Tensor", adv_val: List) -> None:
        """
        Update the specified attack ordering.

        :param images: A tensor of a batch of original inputs to be attacked.
        :param labels: A tensor of a batch of the original labels to be predicted.
        :param adv_val: Optional; A list of a batch of current attack parameters.
        """
        import torch
        import torch.nn.functional as F

        def hungarian(matrix_batch):
            sol = torch.tensor([-i for i in range(1, matrix_batch.shape[0] + 1)], dtype=torch.int32)
            for i in range(matrix_batch.shape[0]):
                topk = 1
                sol[i] = torch.topk(matrix_batch[i], topk)[1][topk - 1]
                while sol.shape != torch.unique(sol).shape:
                    topk = topk + 1
                    sol[i] = torch.topk(matrix_batch[i], topk)[1][topk - 1]
            return sol

        def sinkhorn_normalization(ori_dsm, n_iters=20):
            for _ in range(n_iters):
                ori_dsm /= ori_dsm.sum(dim=0, keepdim=True)
                ori_dsm /= ori_dsm.sum(dim=1, keepdim=True)
            return ori_dsm

        if self.attack_order == "fixed":
            if self.curr_seq.sum() == 0:
                self.fixed_order = tuple(self.enabled_attack.index(i) for i in self.fixed_order)
                self.curr_seq = torch.tensor(self.fixed_order, device=self.device)
        elif self.attack_order == "random":
            self.curr_seq = torch.randperm(self.seq_num)
        elif self.attack_order == "scheduled":
            if self.curr_seq.sum() == 0:
                self.curr_dsm = sinkhorn_normalization(torch.rand((self.seq_num, self.seq_num)))
                self.curr_seq = hungarian(self.curr_dsm)
            self.curr_dsm = self.curr_dsm.detach().requires_grad_()
            adv_img = images.clone().detach().requires_grad_()
            original_iter_num = self.max_inner_iter
            self.max_inner_iter = 3
            self._is_scheduling = True
            for tdx in range(self.seq_num):
                prev_img = adv_img.clone()
                adv_img = torch.zeros_like(adv_img)
                for idx in range(self.seq_num):
                    _adv_img, _ = self.attack_dict[idx](prev_img, adv_val[idx], labels)
                    adv_img = adv_img + self.curr_dsm[tdx][idx] * _adv_img
            self._is_scheduling = False
            self.max_inner_iter = original_iter_num
            outputs = self.model(adv_img)
            with torch.enable_grad():
                cost = F.cross_entropy(outputs, labels)

            dsm_grad = torch.autograd.grad(cost, self.curr_dsm)[0]

            prev_seq = self.curr_seq.clone()
            dsm_noise = torch.zeros_like(self.curr_dsm)
            while torch.equal(prev_seq, self.curr_seq):
                self.curr_dsm = sinkhorn_normalization(torch.exp(self.curr_dsm + dsm_grad + dsm_noise).detach())
                self.curr_seq = hungarian(self.curr_dsm.detach())
                dsm_noise = (torch.randn_like(self.curr_dsm) + 1) * 2  # Escaping local optimum
        else:
            raise ValueError()

    def caa_attack(self, images: "torch.Tensor", labels: "torch.Tensor") -> "torch.Tensor":
        """
        The main algorithm to generate the adversarial examples for composite adversarial attack.

        :param images: A tensor of a batch of original inputs to be attacked.
        :param labels: A tensor of a batch of the original labels to be predicted.
        :return: The perturbed data.
        """
        import torch

        adv_img = images.detach().clone()
        adv_val_saved = torch.zeros((self.seq_num, self.batch_size), device=self.device)
        adv_val = [self.adv_val_space[idx] for idx in range(self.seq_num)]

        if self.is_attacked.sum() > 0:
            for att_id in range(self.seq_num):
                if att_id == self.linf_idx:
                    continue
                adv_val[att_id].detach()
                adv_val[att_id][self.is_attacked] = adv_val_saved[att_id][self.is_attacked]
                adv_val[att_id].requires_grad_()

        for _ in range(self.max_iter):
            self.update_attack_order(images, labels, adv_val)

            adv_img = adv_img.detach().clone()
            self.is_not_attacked = torch.logical_not(self.is_attacked)
            adv_img[self.is_not_attacked] = images[self.is_not_attacked].clone()
            adv_img.requires_grad = True

            for tdx in range(self.seq_num):
                idx = self.curr_seq[tdx]
                adv_img, adv_val_updated = self.attack_dict[idx](adv_img, adv_val[idx], labels)  # type: ignore
                if idx != self.linf_idx:
                    adv_val[idx] = adv_val_updated

            outputs = self.model(adv_img)
            cur_pred = outputs.max(1, keepdim=True)[1].squeeze()
            self.is_attacked = torch.logical_or(self.is_attacked, cur_pred != labels.max(1, keepdim=True)[1].squeeze())

            if self.is_attacked.sum() > 0:
                for att_id in range(self.seq_num):
                    if att_id == self.linf_idx:
                        continue
                    adv_val_saved[att_id][self.is_attacked] = adv_val[att_id][self.is_attacked].detach()

            if self.is_attacked.sum() == self.batch_size:
                break

        return adv_img



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the following white-box attacks related to PE malware crafting:
    1) Append based attacks (example paper link: https://arxiv.org/abs/1810.08280)
    2) Section insertion attacks (example paper link: https://arxiv.org/abs/2008.07125)
    3) Slack manipulation attacks (example paper link: https://arxiv.org/abs/1810.08280)
    4) DOS Header Attacks (example paper link: https://arxiv.org/abs/1901.03583)
"""

import json
import random
import logging
from typing import Optional, Union, Tuple, List, Dict, TYPE_CHECKING
from tqdm.auto import trange

import numpy as np

from art.estimators.estimator import BaseEstimator, NeuralNetworkMixin
from art.estimators.classification.classifier import ClassifierMixin
from art.attacks.attack import EvasionAttack

if TYPE_CHECKING:
    # pylint: disable=C0412
    import tensorflow as tf
    from art.utils import CLASSIFIER_NEURALNETWORK_TYPE

logger = logging.getLogger(__name__)


class MalwareGDTensorFlow(EvasionAttack):
    """
    Implementation of the following white-box attacks related to PE malware crafting:
        1) Append based attacks (example paper link: https://arxiv.org/abs/1810.08280)
        2) Section insertion attacks (example paper link: https://arxiv.org/abs/2008.07125)
        3) Slack manipulation attacks (example paper link: https://arxiv.org/abs/1810.08280)
        4) DOS Header Attacks (example paper link: https://arxiv.org/abs/1901.03583)
    """

    attack_params = EvasionAttack.attack_params + [
        "embedding_weights",
        "param_dic",
        "num_of_iterations",
        "l_0",
        "l_r",
        "use_sign",
        "verbose",
    ]

    _estimator_requirements = (BaseEstimator, NeuralNetworkMixin, ClassifierMixin)

    def __init__(
        self,
        classifier: "CLASSIFIER_NEURALNETWORK_TYPE",
        embedding_weights: np.ndarray,
        param_dic: Dict[str, int],
        num_of_iterations: int = 10,
        l_0: Union[float, int] = 0.1,
        l_r: float = 1.0,
        use_sign: bool = False,
        verbose: bool = False,
    ) -> None:
        """
        :param classifier: A trained classifier that takes in the PE embeddings to make a prediction.
        :param embedding_weights: Weights for the embedding layer
        :param param_dic: A dictionary specifying some MalConv parameters.
                          'maxlen': the input size to the MalConv model
                          'input_dim': the number of discrete values, normally 257.
                          'embedding_size': size of the embedding layer. Default 8.
        :param num_of_iterations: The number of iterations to apply.
        :param l_0: l_0 bound for the attack. If less than 1 it is interpreted as a fraction of the file size.
                    If larger than 1 it is interpreted as the total number of permissible features to change.
        :param l_r: Learning rate for the optimisation
        :param use_sign: If we want to use the sign of the gradient, rather than the gradient itself.
        :param verbose: Show progress bars.
        """
        super().__init__(estimator=classifier)
        self.param_dic = param_dic
        self.embedding_weights = embedding_weights
        self.l_0 = l_0
        self.l_r = l_r
        self.use_sign = use_sign
        self.total_perturbation: np.ndarray = np.zeros(shape=(1, 1))
        self.num_of_iterations = num_of_iterations
        self.verbose = verbose
        self._check_params()

        self.embedding_weights = self.embedding_weights.astype("float32")

    def _check_params(self) -> None:
        if not isinstance(self.param_dic, dict):
            raise ValueError(
                "A param_dic should be provided with the following keys/value pairs: "
                "'maxlen': the input size to the MalConv model"
                "'input_dim': the number of discrete values. Normally 257."
                "'embedding_size': size of the embedding layer. Normally 8."
            )

        if not isinstance(self.embedding_weights, np.ndarray):
            raise ValueError("The weights for the embedding layer should be given as a numpy array.")

        if not isinstance(self.l_0, (int, float)) or self.l_0 < 0:
            raise ValueError(
                "The l0 bound should be greater or equal to 0. Further, it should be provided as an "
                "integer specifying the total number of features to perturb, or a a float representing "
                "the fraction of the total number of features of the original file we can perturb."
            )

        if not isinstance(self.l_r, (int, float)) or self.l_r < 0:
            raise ValueError("The learning rate should be a float or integer greater than zero.")

        if not isinstance(self.use_sign, bool):
            raise ValueError("Whether to use the sign of the gradient should be a True/False bool.")

        if not isinstance(self.num_of_iterations, int) or self.num_of_iterations < 0:
            raise ValueError("The number of iterations must be an integer greater than zero.")

        if not isinstance(self.verbose, bool):
            raise ValueError("The verbosity level should be a True/False bool.")

    @staticmethod
    def initialise_sample(
        x: np.ndarray,
        y: np.ndarray,
        sample_sizes: np.ndarray,
        perturbation_size: np.ndarray,
        perturb_sizes: Optional[List[List[int]]],
        perturb_starts: Optional[List[List[int]]],
    ) -> np.ndarray:
        """
        Randomly append bytes at the end of the malware to initialise it, or if perturbation regions are provided,
        perturb those.

        :param x: Array with input data.
        :param y: Labels, after having been adjusted to account for malware which cannot support the full l0 budget.
        :param sample_sizes: The size of the original file, before it was padded to the input size required by MalConv
        :param perturbation_size: Size of the perturbations in L0 terms to put at end of file
        :param perturb_sizes: List of length batch size, each element is in itself a list containing the size
                              of the allowable perturbation region
        :param perturb_starts: List of length batch size, each element is in itself a list containing the start
                               of perturbation region.
        :return x: Array with features to be perturbed set to a random value.
        """
        for j in range(len(x)):
            if y[j] == 1:
                if perturb_sizes is not None and perturb_starts is not None:
                    for size, start in zip(perturb_sizes[j], perturb_starts[j]):
                        x[j, start : start + size] = np.random.randint(low=0, high=256, size=(1, size))
                x[j, sample_sizes[j] : sample_sizes[j] + perturbation_size[j]] = np.random.randint(
                    low=0, high=256, size=(1, perturbation_size[j])
                )

        return x

    def check_valid_size(
        self,
        y: np.ndarray,
        sample_sizes: np.ndarray,
        append_perturbation_size: np.ndarray,
    ) -> np.ndarray:
        """
        Checks that we can append the l0 perturbation to the malware sample and not exceed the
        maximum file size. A new label vector with just the valid files indicated is created.

        :param y: Labels.
        :param sample_sizes: The size of the original file, before it was padded to the input size required by MalConv.
        :param append_perturbation_size: Size of the perturbations in L0 terms to put at end of file.
        :return adv_label_vector: Labels which indicate which malware samples have enough free features to
                                  accommodate all the adversarial perturbation.
        """

        adv_label_vector = np.zeros_like(y)
        for i, label in enumerate(y):
            if label == 1:
                if sample_sizes[i] + append_perturbation_size[i] <= self.param_dic["maxlen"]:
                    adv_label_vector[i] = 1
                    logger.info("size to append on sample %d is %d", i, append_perturbation_size[i])

        return adv_label_vector

    def generate_mask(
        self,
        x: np.ndarray,
        y: np.ndarray,
        sample_sizes: np.ndarray,
        perturbation_size: np.ndarray,
        perturb_sizes: Optional[List[List[int]]],
        perturb_starts: Optional[List[List[int]]],
    ) -> "tf.Tensor":
        """
        Makes a mask to apply to the gradients to control which samples in the batch are perturbed.

        :param x: Array with input data.
        :param y: Labels to make sure the benign files are zero masked.
        :param sample_sizes: The size of the original file, before it was padded to the input size required by MalConv
        :param perturbation_size: Size of the perturbations in L0 terms to put at end of file
        :param perturb_sizes: List of length batch size, each element is in itself a list containing the size
                              of the allowable perturbation region
        :param perturb_starts: List of length batch size, each element is in itself a list containing the start
                               of perturbation region.
        :return mask: Array with 1s on the features we will modify on this batch and 0s elsewhere.
        """
        import tensorflow as tf

        mask = np.zeros_like(x)
        for i in range(len(x)):
            if y[i] == 1:
                # if no section information was provided, append perturbation at the end of the file.
                if perturb_sizes is None and perturb_starts is None:
                    mask[i, sample_sizes[i] : sample_sizes[i] + perturbation_size[i]] = 1
                elif perturb_sizes is not None and perturb_starts is not None:
                    sample_perturb_sizes = perturb_sizes[i]
                    sample_perturb_starts = perturb_starts[i]

                    for size, start in zip(sample_perturb_sizes, sample_perturb_starts):
                        mask[i, start : start + size] = 1
                    mask[i, sample_sizes[i] : sample_sizes[i] + perturbation_size[i]] = 1
                else:  # pragma: no cover
                    raise ValueError(
                        "either both start and size of perturbation regions are supplied or neither is supplied"
                    )

                # sanity check that the total number of features marked on the mask
                # does not exceed the total perturbation budget
                assert np.sum(mask[i]) == self.total_perturbation[i]
        mask = np.expand_dims(mask, axis=-1)
        expanded_masks = []
        for _ in range(self.param_dic["embedding_size"]):
            expanded_masks.append(mask)
        # repeat it so that it matches the 8 dimensional embedding layer
        expanded_masks = np.concatenate(expanded_masks, axis=-1)
        expanded_masks = tf.convert_to_tensor(expanded_masks)
        expanded_masks = tf.cast(expanded_masks, dtype="float32")  # pylint: disable=E1123 disable=E1120
        return expanded_masks

    def update_embeddings(self, embeddings: "tf.Tensor", gradients: "tf.Tensor", mask: "tf.Tensor") -> "tf.Tensor":
        """
        Update embeddings.

        :param embeddings: Embeddings produced by the data from passing it through the first embedding layer of MalConv
        :param gradients: Gradients to update the embeddings
        :param mask: Tensor with 1s on the embeddings we modify, 0s elsewhere.
        :return embeddings: Updated embeddings wrt the adversarial objective.
        """
        import tensorflow as tf

        if self.use_sign:
            gradients = tf.sign(gradients)
        embeddings = embeddings + self.l_r * gradients * mask
        return embeddings

    def get_adv_malware(
        self,
        embeddings: "tf.Tensor",
        data: np.ndarray,
        labels: np.ndarray,
        fsize: np.ndarray,
        perturbation_size: np.ndarray,
        perturb_sizes: Optional[List[List[int]]] = None,
        perturb_starts: Optional[List[List[int]]] = None,
    ) -> np.ndarray:
        """
        Project the adversarial example back though the closest l2 vector.

        :embeddings: Adversarially optimised embeddings
        :labels: Labels for the data
        :fsize: Size of the original malware
        :data: Original data in the feature space
        :perturbation_size: Size of the l0 attack to append (if any).
        :perturb_sizes: List, with each element itself being a list of the start positions of a
                        perturbation regions in a sample
        :perturb_starts: List, with each element itself being a list of the start positions of a
                         start of the perturbation regions in a sample

        :return data: Numpy array with valid data samples.
        """
        import tensorflow as tf

        for i, label in enumerate(labels):
            if label == 1:
                total_diff = 0
                m = tf.constant([1, self.param_dic["input_dim"], 1], tf.int32)

                if perturb_sizes is not None and perturb_starts is not None:
                    for size, start in zip(perturb_sizes[i], perturb_starts[i]):
                        expanded = tf.tile(tf.expand_dims(embeddings[i, start : start + size, :], axis=1), m)
                        diff = tf.norm((expanded - self.embedding_weights), axis=-1)
                        diff = tf.math.argmin(diff, axis=-1)
                        data[i, start : start + size] = diff
                        total_diff += len(diff)
                expanded = tf.tile(
                    tf.expand_dims(embeddings[i, fsize[i] : fsize[i] + perturbation_size[i], :], axis=1), m
                )
                diff = tf.norm((expanded - self.embedding_weights), axis=-1)
                diff = tf.math.argmin(diff, axis=-1)
                data[i, fsize[i] : fsize[i] + perturbation_size[i]] = diff
                total_diff += len(diff)

                # sanity check that the total number of perturbed features does
                # not exceed the total perturbation budget
                assert total_diff == self.total_perturbation[i]
        return data

    @staticmethod
    def pull_out_adversarial_malware(
        x: np.ndarray,
        y: np.ndarray,
        sample_sizes: np.ndarray,
        initial_dtype: np.dtype,
        input_perturb_sizes: Optional[List[List[int]]] = None,
        input_perturb_starts: Optional[List[List[int]]] = None,
    ) -> Union[
        Tuple[np.ndarray, np.ndarray, np.ndarray],
        Tuple[np.ndarray, np.ndarray, np.ndarray, List[List[int]], List[List[int]]],
    ]:
        """
        Fetches the malware from the data

        :param x: Batch of data which will contain a mix of adversarial examples and unperturbed data.
        :param y: Labels indicating which are valid adversarial examples or not.
        :param initial_dtype: Data can be given in a few formats (uin16, float, etc.) so use initial_dtype
                              to make the returned sample match the original.
        :param sample_sizes: Size of the original data files
        :param input_perturb_sizes: List of length batch size, each element is in itself a list containing
                                    the size of the allowable perturbation region
        :param input_perturb_starts: List of length batch size, each element is in itself a list containing
                                     the start of perturbation region.

        :return adv_x: array composed of only the data that we can make valid adversarial examples from.
        :return adv_y: labels, all ones.
        """
        num_of_malware_samples = int(np.sum(y))

        # make array and allocate, much faster than appending to list and converting
        adv_x = np.zeros((num_of_malware_samples, x.shape[1]), dtype=initial_dtype)
        adv_y = np.ones((num_of_malware_samples, 1))

        adv_sample_sizes = np.zeros((num_of_malware_samples,), dtype=int)
        perturb_sizes = []
        perturb_starts = []

        j = 0
        for i, label in enumerate(y):
            if label == 1:
                adv_x[j] = x[i]
                adv_sample_sizes[j] = int(sample_sizes[i])
                j += 1
                if input_perturb_sizes is not None and input_perturb_starts is not None:
                    perturb_sizes.append(input_perturb_sizes[i])
                    perturb_starts.append(input_perturb_starts[i])

        if input_perturb_sizes is not None and input_perturb_starts is not None:
            return adv_x, adv_y, adv_sample_sizes, perturb_starts, perturb_sizes

        return adv_x, adv_y, adv_sample_sizes

    def compute_perturbation_regions(
        self,
        input_perturbation_size: np.ndarray,
        input_perturb_sizes: List[List[int]],
        automatically_append: bool,
    ) -> Tuple[np.ndarray, List[List[int]]]:
        """
        Based on the l0 budget and the provided allowable perturbation regions we iteratively mark regions of the PE
        file for modification until we exhaust our budget.

        :param input_perturb_sizes: The size of the regions we can perturb.
        :param input_perturbation_size: The total amount of perturbation allowed on a specific sample.
        :param automatically_append: If we want to automatically append unused perturbation on the end of the malware.
        :return perturbation_size: Remaining perturbation (if any)
        :return perturb_sizes: Potentially adjusted sizes of the locations in the PE file we can perturb.
        """
        perturb_sizes = input_perturb_sizes.copy()
        perturbation_size = input_perturbation_size.copy()

        # reduce the perturbation we add at the end if we have perturbation to add in the malware itself and reduce the
        # sizes given in input_perturb_sizes if it exceeds the total input_perturbation_size
        for i, section_sizes in enumerate(perturb_sizes):
            for j, size in enumerate(section_sizes):
                if perturbation_size[i] - size >= 0:
                    logger.info("on sample %d allocate %d in perturb region", i, size)
                    perturbation_size[i] = perturbation_size[i] - size
                else:  # run out of l0 budget.
                    excess = np.abs(perturbation_size[i] - size)  # amount of overspill
                    perturbation_size[i] = perturbation_size[i] - size
                    section_sizes[j] = size - excess
                    # update the perturb sizes

                    logger.info("on sample %d ran out of l0, update to %d from %d", i, section_sizes[j], size)

                    perturb_sizes[i] = section_sizes
                    perturbation_size[i] = 0
        perturbation_size = np.where(perturbation_size < 0, 0, perturbation_size)

        if not automatically_append:
            # if we do not want to automatically append then set the remaining perturbation to 0
            perturbation_size = np.zeros_like(perturbation_size)
            # total_perturbation is now sample dependant so reassign.
            total_perturbation = np.zeros_like(perturbation_size)
            for i in range(len(perturbation_size)):
                total_perturbation[i] = np.sum(perturb_sizes[i])
            self.total_perturbation = total_perturbation
        return perturbation_size, perturb_sizes

    def pull_out_valid_samples(
        self,
        x: np.ndarray,
        y: np.ndarray,
        sample_sizes: np.ndarray,
        automatically_append: bool = True,
        perturb_sizes: Optional[List[List[int]]] = None,
        perturb_starts: Optional[List[List[int]]] = None,
    ) -> Union[
        Tuple[np.ndarray, np.ndarray, np.ndarray],
        Tuple[np.ndarray, np.ndarray, np.ndarray, List[List[int]], List[List[int]]],
    ]:
        """
        Filters the input data for samples that can be made adversarial.

        :param x: Array with input data.
        :param y: Labels to make sure the benign files are zero masked.
        :param sample_sizes: The size of the original file, before it was padded to the input size required by MalConv
        :param automatically_append: Whether to automatically append extra spare perturbation at the end of the file.
        :param perturb_sizes: List of length batch size, each element is in itself a list containing
                              the size of the allowable perturbation region
        :param perturb_starts: List of length batch size, each element is in itself a list containing
                               the start of perturbation region.

        """
        initial_dtype = x.dtype

        perturbation_size = np.zeros(len(sample_sizes), dtype=int)
        for i, sample_size in enumerate(sample_sizes):
            if self.l_0 < 1:  # l0 is a fraction of the filesize
                perturbation_size[i] = int(sample_size * self.l_0)
            else:  # or l0 is interpreted as total perturbation size
                perturbation_size[i] = int(self.l_0)

        if perturb_sizes is not None and perturb_starts is not None:
            perturbation_size, perturb_sizes = self.compute_perturbation_regions(
                perturbation_size, perturb_sizes, automatically_append
            )

        y = self.check_valid_size(y, sample_sizes, perturbation_size)
        if perturb_sizes is not None and perturb_starts is not None:
            return self.pull_out_adversarial_malware(x, y, sample_sizes, initial_dtype, perturb_sizes, perturb_starts)

        return self.pull_out_adversarial_malware(x, y, sample_sizes, initial_dtype)

    def generate(  # pylint: disable=W0221
        self,
        x: np.ndarray,
        y: Optional[np.ndarray] = None,
        sample_sizes: Optional[np.ndarray] = None,
        automatically_append: bool = True,
        verify_input_data: bool = True,
        perturb_sizes: Optional[List[List[int]]] = None,
        perturb_starts: Optional[List[List[int]]] = None,
        **kwargs,
    ) -> np.ndarray:
        """
        Generates the adversarial examples. x needs to be composed of valid files by default which can support the
        adversarial perturbation and so are malicious and can support the assigned L0 budget. They can be obtained by
        using `pull_out_valid_samples` on the data.

        This check on the input data can be over-ridden by toggling the flag verify_input_data
        This will result in only the data which can be made adversarial being perturbed and so the resulting batch will
        be a mixture of adversarial and unperturbed data.

        To assign the L0 budget we go through each list in perturb_sizes and perturb_starts in order, and
        assign the budget based on the sizes given until the l0 budget is exhausted.

        After all the regions marked in perturb_sizes and perturb_starts have been assigned and automatically_append is
        set to true and remaining l0 perturbation the extra perturbation is added at the end in an append style attack.

        :param x: An array with input data.
        :param y: (N, 1) binary labels to make sure the benign files are zero masked.
        :param sample_sizes: The size of the original file, before it was padded to the input size required by MalConv
        :param automatically_append: Whether to automatically append extra spare perturbation at the end of the file.
        :param verify_input_data: If to check that all the data supplied is valid for adversarial perturbations.
        :param perturb_sizes: A list of length batch size, each element is in itself a list containing
                              the size of the allowable perturbation region
        :param perturb_starts: A list of length batch size, each element is in itself a list containing
                               the start of perturbation region.
        :return x: our adversarial examples.
        """
        import tensorflow as tf

        # make copy so original data is not modified.
        adv_x = x.copy()
        if sample_sizes is None:  # pragma: no cover
            raise ValueError("The size of the original files needs to be supplied")
        if y is None:  # pragma: no cover
            raise ValueError("Labels need to be provided so we only modify the malware")

        # check that the dimensions all match
        assert len(adv_x) == len(y)
        assert len(y) == len(sample_sizes)
        if perturb_sizes is not None:
            assert len(y) == len(perturb_sizes)
        if perturb_starts is not None:
            assert len(y) == len(perturb_starts)

        # check that if perturb_starts is provided perturb_sizes is also provided and vise versa
        if perturb_starts is not None:
            assert perturb_sizes is not None
        if perturb_sizes is not None:
            assert perturb_starts is not None

        # if we do not automatically append then make sure that we have supplied
        # start and end positions for the perturbation.
        if not automatically_append:
            assert perturb_sizes is not None
            assert perturb_starts is not None

        perturbation_size = np.zeros(len(sample_sizes), dtype=int)
        for i, sample_size in enumerate(sample_sizes):
            if self.l_0 < 1:  # l0 is a fraction of the filesize
                perturbation_size[i] = int(sample_size * self.l_0)
            else:  # or l0 is interpreted as total perturbation size
                perturbation_size[i] = int(self.l_0)
        self.total_perturbation = np.copy(perturbation_size)

        if perturb_sizes is not None and perturb_starts is not None:
            perturbation_size, perturb_sizes = self.compute_perturbation_regions(
                perturbation_size, perturb_sizes, automatically_append
            )

        y = self.check_valid_size(y, sample_sizes, perturbation_size)

        if verify_input_data:
            if np.sum(y) != len(y):
                raise ValueError(  # pragma: no cover
                    f"{len(y) - np.sum(y)} invalid samples found in batch which cannot support the assigned "
                    f"perturbation or are benign To filter for samples that can be processed use "
                    f"pull_out_valid_samples on the samples. Checking can be disabled by using verify_input_data"
                )

        adv_x = self.initialise_sample(
            adv_x, y, sample_sizes, perturbation_size, perturb_sizes=perturb_sizes, perturb_starts=perturb_starts
        )

        mask = self.generate_mask(
            adv_x, y, sample_sizes, perturbation_size, perturb_sizes=perturb_sizes, perturb_starts=perturb_starts
        )

        embeddings = tf.nn.embedding_lookup(params=self.embedding_weights, ids=adv_x.astype("int32"))

        for _ in trange(self.num_of_iterations, desc="PE Adv. Malware", disable=not self.verbose):
            gradients = self.estimator.class_gradient(embeddings, label=0)
            # go from (bsize x 1 x features x embedding size) -> (bsize x features x embedding size) in a
            # framework-agnostic manner.
            gradients = gradients[:, 0, :, :]
            gradients = -1 * gradients
            embeddings = self.update_embeddings(embeddings, gradients, mask)

        adv_x = self.get_adv_malware(
            embeddings=embeddings,
            data=adv_x,
            labels=y,
            fsize=sample_sizes,
            perturbation_size=perturbation_size,
            perturb_sizes=perturb_sizes,
            perturb_starts=perturb_starts,
        )

        return adv_x

    @staticmethod
    def process_file(
        filepath: str, padding_char: int = 256, maxlen: int = 2 ** 20
    ) -> Tuple[np.ndarray, int]:  # pragma: no cover

        """
        Go from raw file to numpy array.

        :param filepath: Path to the file we convert to a numpy array
        :param padding_char: The char to use to pad the input if it is shorter than maxlen
        :param maxlen: Maximum size of the file processed by the model. Currently set to 1MB
        :return data: A numpy array of the PE file
        :return size_of_original_file: Size of the PE file
        """
        with open(filepath, "rb") as file:
            open_file = file.read()
        size_of_original_file = len(open_file)

        data = np.ones((maxlen,), dtype=np.uint16) * padding_char
        selected_bytes = np.frombuffer(open_file[:maxlen], dtype=np.uint8)
        data[: len(selected_bytes)] = selected_bytes

        return data, size_of_original_file

    @staticmethod
    def get_peinfo(
        filepath: str, save_to_json_path: Optional[str] = None
    ) -> Tuple[List[int], List[int]]:  # pragma: no cover
        """
        Given a PE file we extract out the section information to determine the slack regions in the file.
        We return two lists 1) with the start location of the slack regions and 2) with the size of the slack region.
        We are using the lief library (https://github.com/lief-project/LIEF) to manipulate the PE file.

        :param filepath: Path to file we want to analyse with pedump and get the section information.
        :param save_to_json_path: (Optional) if we want to save the results of pedump to a json file, provide the path.
        :return start_of_slack: A list with the slack starts
        :return size_of_slack: A list with the slack start positions
        """
        import lief  # pylint: disable=C0415

        start_of_slack = []
        size_of_slack = []

        cleaned_dump = {}

        binary_load = lief.parse(filepath)  # pylint: disable=I1101
        if binary_load is not None:
            binary = binary_load
        else:
            raise ValueError("Failed to load binary.")

        for section in binary.sections:
            section_info = {}
            slack = section.sizeof_raw_data - section.virtual_size  # type: ignore
            section_info["PointerToRawData"] = section.pointerto_raw_data  # type: ignore
            section_info["VirtualAddress"] = section.virtual_size  # type: ignore
            section_info["SizeOfRawData"] = section.sizeof_raw_data  # type: ignore
            cleaned_dump[section.name] = section_info
            if slack > 0:
                size_of_slack.append(slack)
                start_of_slack.append(section.pointerto_raw_data + section.virtual_size)  # type: ignore

        if save_to_json_path is not None:
            with open(save_to_json_path, "w", encoding="utf8") as outfile:
                json.dump(cleaned_dump, outfile, indent=4, sort_keys=True)

        return start_of_slack, size_of_slack

    def insert_section(
        self,
        datapoint: Union[List[int], str],
        sample_size: Optional[int] = None,
        padding_char: int = 256,
        maxlen: int = 2 ** 20,
        bytes_to_assign: Optional[int] = None,
        verbose: bool = False,
    ) -> Union[
        Tuple[np.ndarray, int, int, int, List[int], List[int]], Tuple[None, None, None, None, None, None]
    ]:  # pragma: no cover
        """
        Create a new section in a PE file that the attacker can perturb to create an adversarial example.
        we are using the lief library (https://github.com/lief-project/LIEF) to manipulate the PE file.

        :param datapoint: either 1) path to file we want to analyse with lief and get the section information.
                          or 2) list of ints that can be processed by lief.

                          If we have already pre-processed the file into a numpy array, we convert it to a form
                          that can be read by lief.
                          eg, if we have it as a numpy array this could be done by:

                          datapoint = datapoint[0:size]  # size is the original size of the malware file
                          datapoint = datapoint.astype('uint8')
                          datapoint = datapoint.tolist()
        :param sample_size: Size of the original datapoint. Only if it is an array and the l0 budget is fractional
        :param padding_char: The char to use to pad the file to be of length maxlen
        :param maxlen: Maximum length of the data that the MalConv model can process
        :param bytes_to_assign: (Optional) how many bytes we wish to specify when inserting a new section.
                                If unspecified the whole l0 budget will be used on a single section.
        :param verbose: lief outputs a lot to the console, particularly if we are processing many files.
                        By default, suppress printing of messages. Can be toggled on/off by True/False
        :return manipulated_data: Executable with section inserted and turned into a numpy array of
                                  the appropriate size
        :return len(manipulated_file): Size of original file
        :return information_on_section.pointerto_raw_data: The start of the inserted section
        :return information_on_section.virtual_size: Size of the inserted section
        :return size_of_slack: Size of slack regions in this executable (including from the section we just inserted)
        :return start_of_slack: Start of slack regions in this executable (including from the section we just inserted)
        """
        # pylint: disable=I1101
        import lief  # pylint: disable=C0415

        if not verbose:
            lief.logging.disable()

        binary_parse = lief.PE.parse(datapoint)
        if binary_parse is not None:
            binary = binary_parse
        else:
            raise ValueError("Failed to load binary.")

        name_in_use = True
        while name_in_use:
            new_section_name = "." + "".join(chr(random.randrange(ord("a"), ord("z"))) for _ in range(5))
            name_in_use = False

            # check if the random name is already by chance in use
            for section in binary.sections:
                if new_section_name == section.name:
                    name_in_use = True

        new_section = lief.PE.Section(new_section_name)

        if bytes_to_assign is None:
            if self.l_0 < 1:  # l0 is a fraction of the file size
                # if it's a filepath we need to get the file size
                if isinstance(datapoint, str):
                    with open(datapoint, "rb") as file:
                        open_file = file.read()
                    sample_size = len(open_file)
                else:
                    if sample_size is None:
                        raise ValueError(
                            "if the data is an array and the l0 budget is fractional "
                            "the sample size must be provided"
                        )

                perturbation_size = int(sample_size * self.l_0)
            else:  # or l0 is interpreted as total perturbation size
                perturbation_size = int(self.l_0)
            new_section.content = [random.randint(0, 255) for _ in range(perturbation_size)]  # type: ignore
        else:
            new_section.content = [random.randint(0, 255) for _ in range(bytes_to_assign)]  # type: ignore

        # we add the new section at the end of the existing sections
        section_end_points = []
        for section in binary.sections:
            section_end_points.append(section.virtual_address + section.size)

        new_section.virtual_address = max(section_end_points)

        binary.add_section(
            new_section,
            random.choice(
                [
                    lief.PE.SECTION_TYPES.BSS,
                    lief.PE.SECTION_TYPES.DATA,
                    lief.PE.SECTION_TYPES.EXPORT,
                    lief.PE.SECTION_TYPES.IDATA,
                    lief.PE.SECTION_TYPES.RELOCATION,
                    lief.PE.SECTION_TYPES.RESOURCE,
                    lief.PE.SECTION_TYPES.TEXT,
                    lief.PE.SECTION_TYPES.TLS_,
                    lief.PE.SECTION_TYPES.UNKNOWN,
                ]
            ),
        )

        information_on_section = binary.get_section(new_section_name)

        size_of_slack = []
        start_of_slack = []
        for section in binary.sections:
            slack = section.sizeof_raw_data - section.virtual_size
            if slack > 0:
                size_of_slack.append(slack)
                start_of_slack.append(section.pointerto_raw_data + section.virtual_size)

        builder = lief.PE.Builder(binary)
        builder.build()

        manipulated_file = np.array(builder.get_build(), dtype=np.uint8)

        manipulated_data = np.ones((maxlen,), dtype=np.uint16) * padding_char

        # Only process files which are less than the max file size supported
        if len(manipulated_file) < maxlen:
            manipulated_data[: len(manipulated_file)] = manipulated_file[:maxlen]

            return (
                manipulated_data,
                len(manipulated_file),
                information_on_section.pointerto_raw_data,
                information_on_section.virtual_size,
                size_of_slack,
                start_of_slack,
            )
        return None, None, None, None, None, None

    @staticmethod
    def get_dos_locations(x: np.ndarray) -> Tuple[List[List[int]], List[List[int]]]:
        """
        We identify the regions in the DOS header which we can perturb adversarially.

        There are a series of "magic numbers" in this method which relate to the structure of the PE file.
        1) mz_offset = 2: the first two bytes of a PE are fixed as MZ.
        2) 0x3C: offset to the pointer to the PE header. The pointer is 4 bytes long.
        3) 0x40: end of the pointer to the PE header.

        :return batch_of_starts: A list of start locations we can perturb.
                                 This will always have the same value of 2 and 64.
        :return batch_of_sizes: Size of the perturbations we can carry out.
        :return batch_of_starts: Start locations which we can perturb.
        """
        batch_of_starts = []
        batch_of_sizes = []
        mz_offset = 2

        for i in range(len(x)):
            size = []
            start = []

            pointer_to_pe_header = x[i, int(0x3C) : int(0x40)].astype(np.uint8)

            # combine 4 unint8 bytes into unit32
            pointer_to_pe_header = (
                pointer_to_pe_header[3] << 24
                | pointer_to_pe_header[2] << 16
                | pointer_to_pe_header[1] << 8
                | pointer_to_pe_header[0]
            )

            size.append(int(0x3C) - mz_offset)
            start.append(mz_offset)

            size.append(int(pointer_to_pe_header) - int(0x40) - 1)
            start.append(int(0x40))

            batch_of_starts.append(start)
            batch_of_sizes.append(size)

        return batch_of_starts, batch_of_sizes



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module creates GANs using the TensorFlow ML Framework
"""
from typing import Tuple, TYPE_CHECKING, Union

import numpy as np
from art.estimators.tensorflow import TensorFlowV2Estimator

if TYPE_CHECKING:
    from art.utils import CLASSIFIER_TYPE, GENERATOR_TYPE
    import tensorflow as tf


class TensorFlowV2GAN(TensorFlowV2Estimator):
    """
    This class implements a GAN with the TensorFlow v2 framework.
    """

    def __init__(
        self,
        generator: "GENERATOR_TYPE",
        discriminator: "CLASSIFIER_TYPE",
        generator_loss=None,
        discriminator_loss=None,
        generator_optimizer_fct=None,
        discriminator_optimizer_fct=None,
    ):
        """
        Initialization of a test TensorFlow v2 GAN

        :param generator: a TensorFlow2 generator
        :param discriminator: a TensorFlow v2 discriminator
        :param generator_loss: the loss function to use for the generator
        :param discriminator_loss: the loss function to use for the discriminator
        :param generator_optimizer_fct: the optimizer function to use for the generator
        :param discriminator_optimizer_fct: the optimizer function to use for the discriminator
        """
        super().__init__(model=None, clip_values=None, channels_first=None)
        self._generator = generator
        self._discriminator_classifier = discriminator
        self._generator_loss = generator_loss
        self._generator_optimizer_fct = generator_optimizer_fct
        self._discriminator_loss = discriminator_loss
        self._discriminator_optimizer_fct = discriminator_optimizer_fct

    def predict(self, x: np.ndarray, batch_size: int = 128, **kwargs) -> np.ndarray:
        """
        Generates a sample.

        :param x: A input seed.
        :param batch_size: The batch size for predictions.
        :return: The generated sample.
        """
        return self.generator.predict(x, batch_size=batch_size, **kwargs)

    @property
    def input_shape(self) -> Tuple[int, int]:
        """
        Return the shape of one input sample.

        :return: Shape of one input sample.
        """
        return 1, 100

    def fit(self, x: np.ndarray, y: np.ndarray, batch_size: int = 128, nb_epochs: int = 20, **kwargs) -> None:
        """
        Creates a generative model

        :param x: the secret backdoor trigger that will produce the target
        :param y: the target to produce when using the trigger
        :param batch_size: batch_size of images used to train generator
        :param nb_epochs: total number of iterations for performing the attack
        """
        import tensorflow as tf

        z_trigger = x
        for _ in range(nb_epochs):
            train_imgs = kwargs.get("images")
            train_set = (
                tf.data.Dataset.from_tensor_slices(train_imgs)
                .shuffle(train_imgs.shape[0])  # type: ignore
                .batch(batch_size)
            )

            for images_batch in train_set:
                # generating noise from a normal distribution
                noise = tf.random.normal([images_batch.shape[0], z_trigger.shape[1]])

                with tf.GradientTape() as gen_tape, tf.GradientTape() as disc_tape:
                    generated_images = self.generator.model(noise, training=True)
                    real_output = self.discriminator.model(images_batch, training=True)  # type: ignore

                    generated_output = self.discriminator.model(generated_images, training=True)  # type: ignore

                    gen_loss = self._generator_loss(generated_output)
                    disc_loss = self._discriminator_loss(real_output, generated_output)

                gradients_of_generator = gen_tape.gradient(gen_loss, self.generator.model.variables)
                gradients_of_discriminator = disc_tape.gradient(
                    disc_loss, self.discriminator.model.variables  # type: ignore
                )

                self.generator_optimizer_fct.apply_gradients(
                    zip(gradients_of_generator, self.estimator.model.trainable_variables)  # type: ignore
                )
                self.discriminator_optimizer_fct.apply_gradients(
                    zip(gradients_of_discriminator, self.discriminator.model.variables)  # type: ignore
                )

    @property
    def generator(self) -> "GENERATOR_TYPE":
        """
        :return: the generator
        """
        return self._generator

    @property
    def discriminator(self) -> "CLASSIFIER_TYPE":
        """
        :return: the discriminator
        """
        return self._discriminator_classifier

    @property
    def generator_loss(self) -> "tf.Tensor":
        """
        :return: the loss fct used for the generator
        """
        return self._generator_loss

    @property
    def generator_optimizer_fct(self) -> "tf.Tensor":
        """
        :return: the optimizer function for the generator
        """
        return self._generator_optimizer_fct

    @property
    def discriminator_loss(self) -> "tf.Tensor":
        """
        :return: the loss fct used for the discriminator
        """
        return self._discriminator_loss

    @property
    def discriminator_optimizer_fct(self) -> "tf.Tensor":
        """
        :return: the optimizer function for the discriminator
        """
        return self._discriminator_optimizer_fct

    def loss_gradient(self, x, y, **kwargs):
        raise NotImplementedError

    def get_activations(
        self, x: np.ndarray, layer: Union[int, str], batch_size: int, framework: bool = False
    ) -> np.ndarray:
        raise NotImplementedError



"""
Interval based certification estimators.
"""
from art.estimators.certification.interval.interval import PyTorchIntervalDense
from art.estimators.certification.interval.interval import PyTorchIntervalConv2D
from art.estimators.certification.interval.interval import PyTorchIntervalReLU
from art.estimators.certification.interval.interval import PyTorchIntervalFlatten
from art.estimators.certification.interval.interval import PyTorchIntervalBounds
from art.estimators.certification.interval.pytorch import PyTorchIBPClassifier
